import functools
import json
import logging
from typing import TYPE_CHECKING, Any

from mlflow.entities.model_registry.prompt_version import PromptVersion

if TYPE_CHECKING:
    from mlflow.entities import DatasetRecord, EvaluationDataset
    from mlflow.genai.scorers.online.entities import OnlineScoringConfig

from opentelemetry.proto.collector.trace.v1.trace_service_pb2 import ExportTraceServiceRequest
from packaging.version import Version

from mlflow.entities import (
    DatasetInput,
    Experiment,
    LoggedModel,
    LoggedModelInput,
    LoggedModelOutput,
    LoggedModelParameter,
    LoggedModelStatus,
    LoggedModelTag,
    Metric,
    Run,
    RunInfo,
    ScorerVersion,
    ViewType,
)
from mlflow.exceptions import MlflowNotImplementedException

# Constants for Databricks API disabled decorator
_DATABRICKS_DATASET_API_NAME = "Evaluation dataset APIs"
_DATABRICKS_DATASET_ALTERNATIVE = "Use the databricks-agents library for dataset operations."
from mlflow.entities.assessment import Assessment, Expectation, Feedback
from mlflow.entities.span import Span
from mlflow.entities.trace import Trace
from mlflow.entities.trace_data import TraceData
from mlflow.entities.trace_info import TraceInfo
from mlflow.entities.trace_info_v2 import TraceInfoV2
from mlflow.entities.trace_location import TraceLocation
from mlflow.entities.trace_metrics import MetricAggregation, MetricDataPoint, MetricViewType
from mlflow.entities.trace_status import TraceStatus
from mlflow.environment_variables import (
    _MLFLOW_CREATE_LOGGED_MODEL_PARAMS_BATCH_SIZE,
    _MLFLOW_DELETE_TRACES_MAX_BATCH_SIZE,
    _MLFLOW_LOG_LOGGED_MODEL_PARAMS_BATCH_SIZE,
    MLFLOW_ASYNC_TRACE_LOGGING_RETRY_TIMEOUT,
)
from mlflow.exceptions import MlflowException
from mlflow.protos import databricks_pb2
from mlflow.protos.databricks_pb2 import INTERNAL_ERROR
from mlflow.protos.service_pb2 import (
    AddDatasetToExperiments,
    BatchGetTraces,
    CalculateTraceFilterCorrelation,
    CreateAssessment,
    CreateDataset,
    CreateExperiment,
    CreateLoggedModel,
    CreateRun,
    DeleteAssessment,
    DeleteDataset,
    DeleteDatasetTag,
    DeleteExperiment,
    DeleteExperimentTag,
    DeleteLoggedModel,
    DeleteLoggedModelTag,
    DeleteRun,
    DeleteScorer,
    DeleteTag,
    DeleteTraces,
    DeleteTraceTag,
    EndTrace,
    FinalizeLoggedModel,
    GetAssessmentRequest,
    GetDataset,
    GetDatasetExperimentIds,
    GetDatasetRecords,
    GetExperiment,
    GetExperimentByName,
    GetLoggedModel,
    GetMetricHistory,
    GetRun,
    GetScorer,
    GetTrace,
    GetTraceInfo,
    GetTraceInfoV3,
    LinkPromptsToTrace,
    LinkTracesToRun,
    ListScorers,
    ListScorerVersions,
    LogBatch,
    LogInputs,
    LogLoggedModelParamsRequest,
    LogMetric,
    LogModel,
    LogOutputs,
    LogParam,
    MlflowService,
    QueryTraceMetrics,
    RegisterScorer,
    RemoveDatasetFromExperiments,
    RestoreExperiment,
    RestoreRun,
    SearchEvaluationDatasets,
    SearchExperiments,
    SearchLoggedModels,
    SearchRuns,
    SearchTraces,
    SearchTracesV3,
    SetDatasetTags,
    SetExperimentTag,
    SetLoggedModelTags,
    SetTag,
    SetTraceTag,
    StartTrace,
    StartTraceV3,
    TraceRequestMetadata,
    TraceTag,
    UpdateAssessment,
    UpdateExperiment,
    UpdateRun,
    UpsertDatasetRecords,
)
from mlflow.store.entities.paged_list import PagedList
from mlflow.store.tracking import MAX_RESULTS_QUERY_TRACE_METRICS, SEARCH_TRACES_DEFAULT_MAX_RESULTS
from mlflow.store.tracking.abstract_store import AbstractStore
from mlflow.store.tracking.gateway.rest_mixin import RestGatewayStoreMixin
from mlflow.tracing.analysis import TraceFilterCorrelationResult
from mlflow.tracing.utils.otlp import MLFLOW_EXPERIMENT_ID_HEADER, OTLP_TRACES_PATH
from mlflow.utils.databricks_utils import databricks_api_disabled
from mlflow.utils.proto_json_utils import message_to_json
from mlflow.utils.rest_utils import (
    _REST_API_PATH_PREFIX,
    _V3_REST_API_PATH_PREFIX,
    _V3_TRACE_REST_API_PATH_PREFIX,
    MlflowHostCreds,
    call_endpoint,
    extract_api_info_for_service,
    get_logged_model_endpoint,
    get_single_assessment_endpoint,
    get_single_trace_endpoint,
    get_trace_tag_endpoint,
    http_request,
    verify_rest_response,
)
from mlflow.utils.validation import _resolve_experiment_ids_and_locations

_logger = logging.getLogger(__name__)


# MRO Note: RestGatewayStoreMixin must be listed before AbstractStore in the inheritance chain.
# AbstractStore inherits from GatewayStoreMixin which defines abstract Gateway methods.
# RestGatewayStoreMixin provides concrete implementations of those methods. For Python's MRO
# to correctly resolve the Gateway methods to RestGatewayStoreMixin's implementations,
# RestGatewayStoreMixin must appear first in the parent class list.
class RestStore(RestGatewayStoreMixin, AbstractStore):
    """
    Client for a remote tracking server accessed via REST API calls

    Args
        get_host_creds: Method to be invoked prior to every REST request to get the
            :py:class:`mlflow.rest_utils.MlflowHostCreds` for the request. Note that this
            is a function so that we can obtain fresh credentials in the case of expiry.
    """

    _METHOD_TO_INFO = extract_api_info_for_service(MlflowService, _REST_API_PATH_PREFIX)
    _V3_METHOD_TO_INFO = extract_api_info_for_service(MlflowService, _V3_REST_API_PATH_PREFIX)

    # Set of v3 APIs - includes Gateway APIs from mixin
    _V3_APIS = RestGatewayStoreMixin._V3_GATEWAY_APIS

    def __init__(self, get_host_creds):
        super().__init__()
        self.get_host_creds = get_host_creds

    @staticmethod
    @functools.lru_cache
    def _get_server_version(host_creds: MlflowHostCreds) -> Version | None:
        """
        Get the MLflow server version with caching.

        Args:
            host_creds: MlflowHostCreds object

        Returns:
            Version object if successful, None if failed to retrieve version.
        """
        try:
            response = http_request(
                host_creds=host_creds,
                endpoint="/version",
                method="GET",
                timeout=3,  # Short timeout to fail fast if server version API isn't available
                # Disable non-DB SDK retries; default retry policy takes minutes, which is too long
                max_retries=0,
                # Approximately disable DB SDK retries (0 is interpreted as 'unspecified', so use 1)
                retry_timeout_seconds=1,
                raise_on_status=True,
            )
            return Version(response.text)
        except Exception as e:
            _logger.debug(f"Failed to retrieve server version: {e}")

        return None

    def _call_endpoint(
        self,
        api,
        json_body=None,
        endpoint=None,
        retry_timeout_seconds=None,
        response_proto=None,
    ):
        # Route v3 APIs to v3 endpoints, all others to v2 endpoints
        method_to_info = self._V3_METHOD_TO_INFO if api in self._V3_APIS else self._METHOD_TO_INFO

        if endpoint:
            # Allow customizing the endpoint for compatibility with dynamic endpoints, such as
            # /mlflow/traces/{trace_id}/info.
            _, method = method_to_info[api]
        else:
            endpoint, method = method_to_info[api]
        response_proto = response_proto or api.Response()
        return call_endpoint(
            self.get_host_creds(),
            endpoint,
            method,
            json_body,
            response_proto,
            retry_timeout_seconds=retry_timeout_seconds,
        )

    def search_experiments(
        self,
        view_type=ViewType.ACTIVE_ONLY,
        max_results=None,
        filter_string=None,
        order_by=None,
        page_token=None,
    ):
        req_body = message_to_json(
            SearchExperiments(
                view_type=view_type,
                max_results=max_results,
                page_token=page_token,
                order_by=order_by,
                filter=filter_string,
            )
        )
        response_proto = self._call_endpoint(SearchExperiments, req_body)
        experiments = [Experiment.from_proto(x) for x in response_proto.experiments]
        token = (
            response_proto.next_page_token if response_proto.HasField("next_page_token") else None
        )
        return PagedList(experiments, token)

    def create_experiment(self, name, artifact_location=None, tags=None):
        """
        Create a new experiment.
        If an experiment with the given name already exists, throws exception.

        Args:
            name: Desired name for an experiment.
            artifact_location: Location to store run artifacts.
            tags: A list of :py:class:`mlflow.entities.ExperimentTag` instances to set for the
                experiment.

        Returns:
            experiment_id for the newly created experiment if successful, else None
        """
        tag_protos = [tag.to_proto() for tag in tags] if tags else []
        req_body = message_to_json(
            CreateExperiment(name=name, artifact_location=artifact_location, tags=tag_protos)
        )
        response_proto = self._call_endpoint(CreateExperiment, req_body)
        return response_proto.experiment_id

    def get_experiment(self, experiment_id):
        """
        Fetch the experiment from the backend store.

        Args:
            experiment_id: String id for the experiment

        Returns:
            A single :py:class:`mlflow.entities.Experiment` object if it exists,
            otherwise raises an Exception.
        """
        req_body = message_to_json(GetExperiment(experiment_id=str(experiment_id)))
        response_proto = self._call_endpoint(GetExperiment, req_body)
        return Experiment.from_proto(response_proto.experiment)

    def delete_experiment(self, experiment_id):
        req_body = message_to_json(DeleteExperiment(experiment_id=str(experiment_id)))
        self._call_endpoint(DeleteExperiment, req_body)

    def restore_experiment(self, experiment_id):
        req_body = message_to_json(RestoreExperiment(experiment_id=str(experiment_id)))
        self._call_endpoint(RestoreExperiment, req_body)

    def rename_experiment(self, experiment_id, new_name):
        req_body = message_to_json(
            UpdateExperiment(experiment_id=str(experiment_id), new_name=new_name)
        )
        self._call_endpoint(UpdateExperiment, req_body)

    def get_run(self, run_id):
        """
        Fetch the run from backend store

        Args:
            run_id: Unique identifier for the run

        Returns:
            A single Run object if it exists, otherwise raises an Exception
        """
        req_body = message_to_json(GetRun(run_uuid=run_id, run_id=run_id))
        response_proto = self._call_endpoint(GetRun, req_body)
        return Run.from_proto(response_proto.run)

    def update_run_info(self, run_id, run_status, end_time, run_name):
        """Updates the metadata of the specified run."""
        req_body = message_to_json(
            UpdateRun(
                run_uuid=run_id,
                run_id=run_id,
                status=run_status,
                end_time=end_time,
                run_name=run_name,
            )
        )
        response_proto = self._call_endpoint(UpdateRun, req_body)
        return RunInfo.from_proto(response_proto.run_info)

    def create_run(self, experiment_id, user_id, start_time, tags, run_name):
        """
        Create a run under the specified experiment ID, setting the run's status to "RUNNING"
        and the start time to the current time.

        Args:
            experiment_id: ID of the experiment for this run.
            user_id: ID of the user launching this run.
            start_time: timestamp of the initialization of the run.
            tags: tags to apply to this run at initialization.
            run_name: Name of this run.

        Returns:
            The created Run object.
        """

        tag_protos = [tag.to_proto() for tag in tags]
        req_body = message_to_json(
            CreateRun(
                experiment_id=str(experiment_id),
                user_id=user_id,
                start_time=start_time,
                tags=tag_protos,
                run_name=run_name,
            )
        )
        response_proto = self._call_endpoint(CreateRun, req_body)
        return Run.from_proto(response_proto.run)

    def start_trace(self, trace_info: TraceInfo) -> TraceInfo:
        """
        Create a new trace using the V3 API format.

        NB: The backend API is named "StartTraceV3" for some internal reason, but actually
        it is supposed to be called at the end of the trace.

        Args:
            trace_info: The TraceInfo object to create in the backend.

        Returns:
            The returned TraceInfo object from the backend.
        """
        # NB: The Databricks backend expects a Trace object, not a TraceInfo object, although
        # it doesn't use the data field at all. Trace data increases the payload size significantly,
        # so we create a Trace object with an empty data field here.
        trace = Trace(info=trace_info, data=TraceData(spans=[]))
        req_body = message_to_json(StartTraceV3(trace=trace.to_proto()))

        try:
            response_proto = self._call_endpoint(
                # NB: _call_endpoint doesn't handle versioning between v2 and v3 endpoint
                # yet, so manually passing the v3 endpoint here.
                StartTraceV3,
                req_body,
                endpoint=_V3_TRACE_REST_API_PATH_PREFIX,
                retry_timeout_seconds=MLFLOW_ASYNC_TRACE_LOGGING_RETRY_TIMEOUT.get(),
            )
            return TraceInfo.from_proto(response_proto.trace.trace_info)
        except MlflowException as e:
            if e.error_code == databricks_pb2.ErrorCode.Name(databricks_pb2.ENDPOINT_NOT_FOUND):
                _logger.debug(
                    "Server does not support StartTraceV3 API yet. Falling back to V2 API."
                )
                return self._create_trace_v2_fallback(trace_info)
            raise

    def _create_trace_v2_fallback(self, trace_info: TraceInfo) -> TraceInfo:
        """
        Create a new trace using the V2 API format. This is a fallback for the case where the
        client is v3 but the tracking server does not support v3 yet(<= 3.2.0).
        """
        trace_info_v2 = self.deprecated_start_trace_v2(
            experiment_id=trace_info.experiment_id,
            timestamp_ms=trace_info.request_time,
            request_metadata=trace_info.trace_metadata,
            tags=trace_info.tags,
        )
        self.deprecated_end_trace_v2(
            request_id=trace_info_v2.request_id,
            timestamp_ms=trace_info.request_time + trace_info.execution_duration,
            status=trace_info.status,
            request_metadata=trace_info.trace_metadata,
            tags=trace_info.tags,
        )
        return trace_info_v2.to_v3()

    def _delete_traces(
        self,
        experiment_id: str,
        max_timestamp_millis: int | None = None,
        max_traces: int | None = None,
        trace_ids: list[str] | None = None,
    ) -> int:
        # If deleting by trace_ids, split into batches to avoid hitting the
        # Databricks server limit
        if trace_ids:
            batch_size = _MLFLOW_DELETE_TRACES_MAX_BATCH_SIZE.get()
            total_deleted = 0
            for i in range(0, len(trace_ids), batch_size):
                batch = trace_ids[i : i + batch_size]
                req_body = message_to_json(
                    DeleteTraces(
                        experiment_id=experiment_id,
                        max_timestamp_millis=max_timestamp_millis,
                        max_traces=max_traces,
                        request_ids=batch,
                    )
                )
                res = self._call_endpoint(DeleteTraces, req_body)
                total_deleted += res.traces_deleted
            return total_deleted
        else:
            req_body = message_to_json(
                DeleteTraces(
                    experiment_id=experiment_id,
                    max_timestamp_millis=max_timestamp_millis,
                    max_traces=max_traces,
                    request_ids=trace_ids,
                )
            )
            res = self._call_endpoint(DeleteTraces, req_body)
            return res.traces_deleted

    def get_trace_info(self, trace_id: str) -> TraceInfo:
        """
        Get the trace matching the `trace_id`.

        Args:
            trace_id: String id of the trace to fetch.

        Returns:
            The fetched Trace object, of type ``mlflow.entities.TraceInfo``.
        """
        trace_v3_req_body = message_to_json(GetTraceInfoV3(trace_id=trace_id))
        trace_v3_endpoint = get_single_trace_endpoint(trace_id)
        try:
            trace_v3_response_proto = self._call_endpoint(
                GetTraceInfoV3, trace_v3_req_body, endpoint=trace_v3_endpoint
            )
            return TraceInfo.from_proto(trace_v3_response_proto.trace.trace_info)
        except MlflowException as e:
            # If the tracking server does not support V3 trace API yet, fallback to V2 API.
            if e.error_code != databricks_pb2.ErrorCode.Name(databricks_pb2.ENDPOINT_NOT_FOUND):
                raise
            _logger.debug("Server does not support GetTraceInfoV3 API yet. Falling back to V2 API.")

        req_body = message_to_json(GetTraceInfo(request_id=trace_id))
        endpoint = get_single_trace_endpoint(trace_id, use_v3=False)
        response_proto = self._call_endpoint(GetTraceInfo, req_body, endpoint=endpoint)
        return TraceInfoV2.from_proto(response_proto.trace_info).to_v3()

    def get_trace(self, trace_id: str, *, allow_partial: bool = False) -> Trace:
        req_body = message_to_json(GetTrace(trace_id=trace_id, allow_partial=allow_partial))
        try:
            response_proto = self._call_endpoint(
                GetTrace, req_body, endpoint=f"{_V3_TRACE_REST_API_PATH_PREFIX}/get"
            )
            return Trace.from_proto(response_proto.trace)
        except MlflowException as e:
            # Note: Old servers might match the route pattern /mlflow/traces/{trace_id} and
            # interpret "get" as the trace_id parameter, we should catch this and raise
            # NotImplementedException for downstream handling.
            if (
                e.error_code
                == databricks_pb2.ErrorCode.Name(databricks_pb2.RESOURCE_DOES_NOT_EXIST)
                and "Trace with ID 'get' not found" in e.message
            ):
                raise MlflowNotImplementedException()
            raise

    def batch_get_traces(self, trace_ids: list[str], location: str | None = None) -> list[Trace]:
        """
        Get a batch of complete traces with spans for given trace ids.

        Args:
            trace_ids: List of trace IDs to fetch.
            location: Location of the trace. Should be None for OSS backend.

        Returns:
            List of Trace objects.
        """
        req_body = message_to_json(BatchGetTraces(trace_ids=trace_ids))
        response_proto = self._call_endpoint(
            BatchGetTraces, req_body, endpoint=f"{_V3_TRACE_REST_API_PATH_PREFIX}/batchGet"
        )
        return [Trace.from_proto(proto) for proto in response_proto.traces]

    def search_traces(
        self,
        experiment_ids: list[str] | None = None,
        filter_string: str | None = None,
        max_results: int = SEARCH_TRACES_DEFAULT_MAX_RESULTS,
        order_by: list[str] | None = None,
        page_token: str | None = None,
        model_id: str | None = None,
        locations: list[str] | None = None,
    ):
        locations = _resolve_experiment_ids_and_locations(experiment_ids, locations)

        if model_id is not None:
            raise MlflowException.invalid_parameter_value(
                "Searching traces by model_id is not supported on the current tracking server.",
            )

        return self._search_traces(
            locations=locations,
            filter_string=filter_string,
            max_results=max_results,
            order_by=order_by,
            page_token=page_token,
        )

    def _search_traces(
        self,
        locations: list[str],
        filter_string: str | None = None,
        max_results: int = SEARCH_TRACES_DEFAULT_MAX_RESULTS,
        order_by: list[str] | None = None,
        page_token: str | None = None,
    ) -> tuple[list[TraceInfo], str | None]:
        # Create trace_locations from experiment_ids for the V3 API
        trace_locations = []
        for exp_id in locations:
            try:
                location = TraceLocation.from_experiment_id(exp_id)
                proto_location = location.to_proto()
                trace_locations.append(proto_location)
            except Exception as e:
                raise MlflowException(
                    f"Invalid experiment ID format: {exp_id}. Error: {e!s}"
                ) from e

        # Create V3 request message using protobuf
        request = SearchTracesV3(
            locations=trace_locations,
            filter=filter_string,
            max_results=max_results,
            order_by=order_by,
            page_token=page_token,
        )

        req_body = message_to_json(request)
        v3_endpoint = f"{_V3_TRACE_REST_API_PATH_PREFIX}/search"

        try:
            response_proto = self._call_endpoint(SearchTracesV3, req_body, v3_endpoint)
        except MlflowException as e:
            if e.error_code == databricks_pb2.ErrorCode.Name(databricks_pb2.ENDPOINT_NOT_FOUND):
                _logger.debug(
                    "Server does not support SearchTracesV3 API yet. Falling back to V2 API."
                )
                response_proto = self._call_endpoint(SearchTraces, req_body)
            else:
                raise

        trace_infos = [TraceInfo.from_proto(t) for t in response_proto.traces]
        return trace_infos, response_proto.next_page_token or None

    def calculate_trace_filter_correlation(
        self,
        experiment_ids: list[str],
        filter_string1: str,
        filter_string2: str,
        base_filter: str | None = None,
    ) -> TraceFilterCorrelationResult:
        """
        Calculate correlation between two trace filter conditions using NPMI.

        Args:
            experiment_ids: List of experiment_ids to search over
            filter_string1: First filter condition in search_traces filter syntax
            filter_string2: Second filter condition in search_traces filter syntax
            base_filter: Optional base filter that both filter1 and filter2 are tested on top of

        Returns:
            TraceFilterCorrelationResult containing NPMI analytics data.
        """
        request = CalculateTraceFilterCorrelation(
            experiment_ids=experiment_ids,
            filter_string1=filter_string1,
            filter_string2=filter_string2,
        )
        if base_filter is not None:
            request.base_filter = base_filter

        req_body = message_to_json(request)
        v3_endpoint = f"{_V3_TRACE_REST_API_PATH_PREFIX}/calculate-filter-correlation"
        response_proto = self._call_endpoint(CalculateTraceFilterCorrelation, req_body, v3_endpoint)
        return TraceFilterCorrelationResult.from_proto(response_proto)

    def set_trace_tag(self, trace_id: str, key: str, value: str):
        """
        Set a tag on the trace with the given trace_id.

        Args:
            trace_id: The ID of the trace.
            key: The string key of the tag.
            value: The string value of the tag.
        """
        # Always use v2 endpoint
        req_body = message_to_json(SetTraceTag(key=key, value=value))
        self._call_endpoint(SetTraceTag, req_body, endpoint=get_trace_tag_endpoint(trace_id))

    def delete_trace_tag(self, trace_id: str, key: str):
        """
        Delete a tag on the trace with the given trace_id.

        Args:
            trace_id: The ID of the trace.
            key: The string key of the tag.
        """
        # Always use v2 endpoint
        req_body = message_to_json(DeleteTraceTag(key=key))
        self._call_endpoint(DeleteTraceTag, req_body, endpoint=get_trace_tag_endpoint(trace_id))

    def query_trace_metrics(
        self,
        experiment_ids: list[str],
        view_type: MetricViewType,
        metric_name: str,
        aggregations: list[MetricAggregation],
        dimensions: list[str] | None = None,
        filters: list[str] | None = None,
        time_interval_seconds: int | None = None,
        start_time_ms: int | None = None,
        end_time_ms: int | None = None,
        max_results: int = MAX_RESULTS_QUERY_TRACE_METRICS,
        page_token: str | None = None,
    ) -> PagedList[MetricDataPoint]:
        max_results = max_results or MAX_RESULTS_QUERY_TRACE_METRICS
        request = QueryTraceMetrics(
            experiment_ids=experiment_ids,
            view_type=view_type.to_proto(),
            metric_name=metric_name,
            aggregations=[agg.to_proto() for agg in aggregations],
            max_results=max_results,
        )
        if dimensions:
            request.dimensions.extend(dimensions)
        if filters:
            request.filters.extend(filters)
        if time_interval_seconds is not None:
            request.time_interval_seconds = time_interval_seconds
        if start_time_ms is not None:
            request.start_time_ms = start_time_ms
        if end_time_ms is not None:
            request.end_time_ms = end_time_ms
        if page_token is not None:
            request.page_token = page_token

        req_body = message_to_json(request)
        endpoint = f"{_V3_TRACE_REST_API_PATH_PREFIX}/metrics"
        response_proto = self._call_endpoint(QueryTraceMetrics, req_body, endpoint)

        data_points = [MetricDataPoint.from_proto(dp) for dp in response_proto.data_points]
        token = response_proto.next_page_token or None
        return PagedList(data_points, token)

    def get_assessment(self, trace_id: str, assessment_id: str) -> Assessment:
        """
        Get an assessment entity from the backend store.
        """
        req_body = message_to_json(
            GetAssessmentRequest(trace_id=trace_id, assessment_id=assessment_id)
        )
        response_proto = self._call_endpoint(
            GetAssessmentRequest,
            req_body,
            endpoint=get_single_assessment_endpoint(trace_id, assessment_id),
        )
        return Assessment.from_proto(response_proto.assessment)

    def create_assessment(self, assessment: Assessment) -> Assessment:
        """
        Create an assessment entity in the backend store.

        Args:
            assessment: The assessment to log (without an assessment_id).

        Returns:
            The created Assessment object.
        """
        req_body = message_to_json(CreateAssessment(assessment=assessment.to_proto()))
        response_proto = self._call_endpoint(
            CreateAssessment,
            req_body,
            endpoint=f"{_V3_TRACE_REST_API_PATH_PREFIX}/{assessment.trace_id}/assessments",
        )
        return Assessment.from_proto(response_proto.assessment)

    def update_assessment(
        self,
        trace_id: str,
        assessment_id: str,
        name: str | None = None,
        expectation: Expectation | None = None,
        feedback: Feedback | None = None,
        rationale: str | None = None,
        metadata: dict[str, str] | None = None,
    ) -> Assessment:
        """
        Update an existing assessment entity in the backend store.

        Args:
            trace_id: The ID of the trace.
            assessment_id: The ID of the assessment to update.
            name: The updated name of the assessment.
            expectation: The updated expectation value of the assessment.
            feedback: The updated feedback value of the assessment.
            rationale: The updated rationale of the feedback. Not applicable for expectations.
            metadata: Additional metadata for the assessment.
        """
        if expectation is not None and feedback is not None:
            raise MlflowException.invalid_parameter_value(
                "Exactly one of `expectation` or `feedback` should be specified."
            )

        update = UpdateAssessment()

        # The assessment object to be sent to the backend (only contains fields to update and IDs)
        assessment = update.assessment
        # Field mask specifies which fields to update.
        mask = update.update_mask

        assessment.assessment_id = assessment_id
        assessment.trace_id = trace_id

        if name is not None:
            assessment.assessment_name = name
            mask.paths.append("assessment_name")
        if expectation is not None:
            assessment.expectation.CopyFrom(expectation.to_proto())
            mask.paths.append("expectation")
        if feedback is not None:
            assessment.feedback.CopyFrom(feedback.to_proto())
            mask.paths.append("feedback")
        if rationale is not None:
            assessment.rationale = rationale
            mask.paths.append("rationale")
        if metadata is not None:
            assessment.metadata.update(metadata)
            mask.paths.append("metadata")

        req_body = message_to_json(update)
        response_proto = self._call_endpoint(
            UpdateAssessment,
            req_body,
            endpoint=get_single_assessment_endpoint(trace_id, assessment_id),
        )
        return Assessment.from_proto(response_proto.assessment)

    def delete_assessment(self, trace_id: str, assessment_id: str):
        """
        Delete an assessment associated with a trace.

        Args:
            trace_id: String ID of the trace.
            assessment_id: String ID of the assessment to delete.
        """
        req_body = message_to_json(DeleteAssessment(trace_id=trace_id, assessment_id=assessment_id))
        self._call_endpoint(
            DeleteAssessment,
            req_body,
            endpoint=get_single_assessment_endpoint(trace_id, assessment_id),
        )

    def log_metric(self, run_id: str, metric: Metric):
        """
        Log a metric for the specified run

        Args:
            run_id: String id for the run
            metric: Metric instance to log
        """
        req_body = message_to_json(
            LogMetric(
                run_uuid=run_id,
                run_id=run_id,
                key=metric.key,
                value=metric.value,
                timestamp=metric.timestamp,
                step=metric.step,
                model_id=metric.model_id,
                dataset_name=metric.dataset_name,
                dataset_digest=metric.dataset_digest,
            )
        )
        self._call_endpoint(LogMetric, req_body)

    def log_param(self, run_id, param):
        """
        Log a param for the specified run

        Args:
            run_id: String id for the run
            param: Param instance to log
        """
        req_body = message_to_json(
            LogParam(run_uuid=run_id, run_id=run_id, key=param.key, value=param.value)
        )
        self._call_endpoint(LogParam, req_body)

    def set_experiment_tag(self, experiment_id, tag):
        """
        Set a tag for the specified experiment

        Args:
            experiment_id: String ID of the experiment
            tag: ExperimentRunTag instance to log
        """
        req_body = message_to_json(
            SetExperimentTag(experiment_id=experiment_id, key=tag.key, value=tag.value)
        )
        self._call_endpoint(SetExperimentTag, req_body)

    def delete_experiment_tag(self, experiment_id, key):
        """
        Delete a tag from the specified experiment

        Args:
            experiment_id: String ID of the experiment
            key: String name of the tag to be deleted
        """
        req_body = message_to_json(DeleteExperimentTag(experiment_id=experiment_id, key=key))
        self._call_endpoint(DeleteExperimentTag, req_body)

    def set_tag(self, run_id, tag):
        """
        Set a tag for the specified run

        Args:
            run_id: String ID of the run
            tag: RunTag instance to log
        """
        req_body = message_to_json(
            SetTag(run_uuid=run_id, run_id=run_id, key=tag.key, value=tag.value)
        )
        self._call_endpoint(SetTag, req_body)

    def delete_tag(self, run_id, key):
        """
        Delete a tag from a run. This is irreversible.

        Args:
            run_id: String ID of the run.
            key: Name of the tag.
        """
        req_body = message_to_json(DeleteTag(run_id=run_id, key=key))
        self._call_endpoint(DeleteTag, req_body)

    def get_metric_history(self, run_id, metric_key, max_results=None, page_token=None):
        """
        Return all logged values for a given metric.

        Args:
            run_id: Unique identifier for run.
            metric_key: Metric name within the run.
            max_results: Maximum number of metric history events (steps) to return per paged
                query. Only supported in 'databricks' backend.
            page_token: A Token specifying the next paginated set of results of metric history.

        Returns:
            A PagedList of :py:class:`mlflow.entities.Metric` entities if a paginated request
            is made by setting ``max_results`` to a value other than ``None``, a List of
            :py:class:`mlflow.entities.Metric` entities if ``max_results`` is None, else, if no
            metrics of the ``metric_key`` have been logged to the ``run_id``, an empty list.
        """
        req_body = message_to_json(
            GetMetricHistory(
                run_uuid=run_id,
                run_id=run_id,
                metric_key=metric_key,
                max_results=max_results,
                page_token=page_token,
            )
        )
        response_proto = self._call_endpoint(GetMetricHistory, req_body)

        metric_history = [Metric.from_proto(metric) for metric in response_proto.metrics]
        return PagedList(metric_history, response_proto.next_page_token or None)

    def _search_runs(
        self, experiment_ids, filter_string, run_view_type, max_results, order_by, page_token
    ):
        experiment_ids = [str(experiment_id) for experiment_id in experiment_ids]
        sr = SearchRuns(
            experiment_ids=experiment_ids,
            filter=filter_string,
            run_view_type=ViewType.to_proto(run_view_type),
            max_results=max_results,
            order_by=order_by,
            page_token=page_token,
        )
        req_body = message_to_json(sr)
        response_proto = self._call_endpoint(SearchRuns, req_body)
        runs = [Run.from_proto(proto_run) for proto_run in response_proto.runs]
        # If next_page_token is not set, we will see it as "". We need to convert this to None.
        next_page_token = None
        if response_proto.next_page_token:
            next_page_token = response_proto.next_page_token
        return runs, next_page_token

    def delete_run(self, run_id):
        req_body = message_to_json(DeleteRun(run_id=run_id))
        self._call_endpoint(DeleteRun, req_body)

    def restore_run(self, run_id):
        req_body = message_to_json(RestoreRun(run_id=run_id))
        self._call_endpoint(RestoreRun, req_body)

    def get_experiment_by_name(self, experiment_name):
        try:
            req_body = message_to_json(GetExperimentByName(experiment_name=experiment_name))
            response_proto = self._call_endpoint(GetExperimentByName, req_body)
            return Experiment.from_proto(response_proto.experiment)
        except MlflowException as e:
            if e.error_code == databricks_pb2.ErrorCode.Name(
                databricks_pb2.RESOURCE_DOES_NOT_EXIST
            ):
                return None
            else:
                raise

    def log_batch(self, run_id, metrics, params, tags):
        metric_protos = [metric.to_proto() for metric in metrics]
        param_protos = [param.to_proto() for param in params]
        tag_protos = [tag.to_proto() for tag in tags]
        req_body = message_to_json(
            LogBatch(metrics=metric_protos, params=param_protos, tags=tag_protos, run_id=run_id)
        )
        self._call_endpoint(LogBatch, req_body)

    def record_logged_model(self, run_id, mlflow_model):
        req_body = message_to_json(
            LogModel(run_id=run_id, model_json=json.dumps(mlflow_model.get_tags_dict()))
        )
        self._call_endpoint(LogModel, req_body)

    def create_logged_model(
        self,
        experiment_id: str,
        name: str | None = None,
        source_run_id: str | None = None,
        tags: list[LoggedModelTag] | None = None,
        params: list[LoggedModelParameter] | None = None,
        model_type: str | None = None,
    ) -> LoggedModel:
        """
        Create a new logged model.

        Args:
            experiment_id: ID of the experiment to which the model belongs.
            name: Name of the model. If not specified, a random name will be generated.
            source_run_id: ID of the run that produced the model.
            tags: Tags to set on the model.
            params: Parameters to set on the model.
            model_type: Type of the model.

        Returns:
            The created model.
        """
        # Include the first 100 params in the initial request
        initial_params = []
        remaining_params = []
        if params:
            initial_batch_size = _MLFLOW_CREATE_LOGGED_MODEL_PARAMS_BATCH_SIZE.get()
            initial_params = params[:initial_batch_size]
            remaining_params = params[initial_batch_size:]

        req_body = message_to_json(
            CreateLoggedModel(
                experiment_id=experiment_id,
                name=name,
                model_type=model_type,
                source_run_id=source_run_id,
                params=[p.to_proto() for p in initial_params],
                tags=[t.to_proto() for t in tags or []],
            )
        )

        response_proto = self._call_endpoint(CreateLoggedModel, req_body)
        model = LoggedModel.from_proto(response_proto.model)

        # Log remaining params if there are any
        if remaining_params:
            self.log_logged_model_params(model_id=model.model_id, params=remaining_params)
            model = self.get_logged_model(model_id=model.model_id)

        return model

    def log_logged_model_params(self, model_id: str, params: list[LoggedModelParameter]) -> None:
        """
        Log parameters for a logged model in batches of 100.

        Args:
            model_id: ID of the model to log parameters for.
            params: List of parameters to log.

        Returns:
            None
        """
        # Process params in batches to avoid exceeding per-request backend limits
        batch_size = _MLFLOW_LOG_LOGGED_MODEL_PARAMS_BATCH_SIZE.get()
        endpoint = get_logged_model_endpoint(model_id)
        for i in range(0, len(params), batch_size):
            batch = params[i : i + batch_size]
            req_body = message_to_json(
                LogLoggedModelParamsRequest(
                    model_id=model_id,
                    params=[p.to_proto() for p in batch],
                )
            )
            self._call_endpoint(
                LogLoggedModelParamsRequest, json_body=req_body, endpoint=f"{endpoint}/params"
            )

    def get_logged_model(self, model_id: str, allow_deleted: bool = False) -> LoggedModel:
        """
        Fetch the logged model with the specified ID.

        Args:
            model_id: ID of the model to fetch.
            allow_deleted: If ``True``, allow fetching logged models in the deleted lifecycle
                stage. Defaults to ``False``.

        Returns:
            The fetched model.
        """
        endpoint = get_logged_model_endpoint(model_id)
        if allow_deleted:
            endpoint = f"{endpoint}?allow_deleted=true"
        response_proto = self._call_endpoint(GetLoggedModel, endpoint=endpoint)
        return LoggedModel.from_proto(response_proto.model)

    def delete_logged_model(self, model_id) -> None:
        request = DeleteLoggedModel(model_id=model_id)
        endpoint = get_logged_model_endpoint(model_id)
        self._call_endpoint(
            DeleteLoggedModel, endpoint=endpoint, json_body=message_to_json(request)
        )

    def search_logged_models(
        self,
        experiment_ids: list[str],
        filter_string: str | None = None,
        datasets: list[dict[str, Any]] | None = None,
        max_results: int | None = None,
        order_by: list[dict[str, Any]] | None = None,
        page_token: str | None = None,
    ) -> PagedList[LoggedModel]:
        """
        Search for logged models that match the specified search criteria.

        Args:
            experiment_ids: List of experiment ids to scope the search.
            filter_string: A search filter string.
            datasets: List of dictionaries to specify datasets on which to apply metrics filters.
                The following fields are supported:

                dataset_name (str): Required. Name of the dataset.
                dataset_digest (str): Optional. Digest of the dataset.
            max_results: Maximum number of logged models desired.
            order_by: List of dictionaries to specify the ordering of the search results.
                The following fields are supported:

                field_name (str): Required. Name of the field to order by, e.g. "metrics.accuracy".
                ascending: (bool): Optional. Whether the order is ascending or not.
                dataset_name: (str): Optional. If ``field_name`` refers to a metric, this field
                    specifies the name of the dataset associated with the metric. Only metrics
                    associated with the specified dataset name will be considered for ordering.
                    This field may only be set if ``field_name`` refers to a metric.
                dataset_digest (str): Optional. If ``field_name`` refers to a metric, this field
                    specifies the digest of the dataset associated with the metric. Only metrics
                    associated with the specified dataset name and digest will be considered for
                    ordering. This field may only be set if ``dataset_name`` is also set.
            page_token: Token specifying the next page of results.

        Returns:
            A :py:class:`PagedList <mlflow.store.entities.PagedList>` of
            :py:class:`LoggedModel <mlflow.entities.LoggedModel>` objects.
        """
        req_body = message_to_json(
            SearchLoggedModels(
                experiment_ids=experiment_ids,
                filter=filter_string,
                datasets=[
                    SearchLoggedModels.Dataset(
                        dataset_name=d["dataset_name"],
                        dataset_digest=d.get("dataset_digest"),
                    )
                    for d in datasets or []
                ],
                max_results=max_results,
                order_by=[
                    SearchLoggedModels.OrderBy(
                        field_name=d["field_name"],
                        ascending=d.get("ascending", True),
                        dataset_name=d.get("dataset_name"),
                        dataset_digest=d.get("dataset_digest"),
                    )
                    for d in order_by or []
                ],
                page_token=page_token,
            )
        )
        response_proto = self._call_endpoint(SearchLoggedModels, req_body)
        models = [LoggedModel.from_proto(x) for x in response_proto.models]
        return PagedList(models, response_proto.next_page_token or None)

    def finalize_logged_model(self, model_id: str, status: LoggedModelStatus) -> LoggedModel:
        """
        Finalize a model by updating its status.

        Args:
            model_id: ID of the model to finalize.
            status: Final status to set on the model.

        Returns:
            The updated model.
        """
        endpoint = get_logged_model_endpoint(model_id)
        json_body = message_to_json(
            FinalizeLoggedModel(model_id=model_id, status=status.to_proto())
        )
        response_proto = self._call_endpoint(
            FinalizeLoggedModel, json_body=json_body, endpoint=endpoint
        )
        return LoggedModel.from_proto(response_proto.model)

    def set_logged_model_tags(self, model_id: str, tags: list[LoggedModelTag]) -> None:
        """
        Set tags on the specified logged model.

        Args:
            model_id: ID of the model.
            tags: Tags to set on the model.

        Returns:
            None
        """
        endpoint = get_logged_model_endpoint(model_id)
        json_body = message_to_json(SetLoggedModelTags(tags=[tag.to_proto() for tag in tags]))
        self._call_endpoint(SetLoggedModelTags, json_body=json_body, endpoint=f"{endpoint}/tags")

    def delete_logged_model_tag(self, model_id: str, key: str) -> None:
        """
        Delete a tag from the specified logged model.

        Args:
            model_id: ID of the model.
            key: Key of the tag to delete.

        Returns:
            The model with the specified tag removed.
        """
        endpoint = get_logged_model_endpoint(model_id)
        self._call_endpoint(DeleteLoggedModelTag, endpoint=f"{endpoint}/tags/{key}")

    def log_inputs(
        self,
        run_id: str,
        datasets: list[DatasetInput] | None = None,
        models: list[LoggedModelInput] | None = None,
    ):
        """
        Log inputs, such as datasets, to the specified run.

        Args:
            run_id: String id for the run
            datasets: List of :py:class:`mlflow.entities.DatasetInput` instances to log
                as inputs to the run.
            models: List of :py:class:`mlflow.entities.LoggedModelInput` instances to log.

        Returns:
            None.
        """
        datasets_protos = [dataset.to_proto() for dataset in datasets or []]
        models_protos = [model.to_proto() for model in models or []]
        req_body = message_to_json(
            LogInputs(
                run_id=run_id,
                datasets=datasets_protos,
                models=models_protos,
            )
        )
        self._call_endpoint(LogInputs, req_body)

    def log_outputs(self, run_id: str, models: list[LoggedModelOutput]):
        """
        Log outputs, such as models, to the specified run.

        Args:
            run_id: String id for the run
            models: List of :py:class:`mlflow.entities.LoggedModelOutput` instances to log
                as outputs of the run.

        Returns:
            None.
        """
        req_body = message_to_json(LogOutputs(run_id=run_id, models=[m.to_proto() for m in models]))
        self._call_endpoint(LogOutputs, req_body)

    ############################################################################################
    # Scorer Management APIs
    ############################################################################################

    def register_scorer(
        self, experiment_id: str, name: str, serialized_scorer: str
    ) -> ScorerVersion:
        """
        Register a scorer for an experiment.

        Args:
            experiment_id: String ID of the experiment.
            name: String name of the scorer.
            serialized_scorer: String containing the serialized scorer data.

        Returns:
            ScorerVersion: The newly registered scorer version object.
        """
        req_body = message_to_json(
            RegisterScorer(
                experiment_id=experiment_id,
                name=name,
                serialized_scorer=serialized_scorer,
            )
        )
        # Scorer APIs are v3.0 endpoints
        response_proto = self._call_endpoint(
            RegisterScorer,
            req_body,
            endpoint="/api/3.0/mlflow/scorers/register",
        )

        return ScorerVersion(
            experiment_id=response_proto.experiment_id,
            scorer_name=response_proto.name,
            scorer_version=response_proto.version,
            serialized_scorer=response_proto.serialized_scorer,
            creation_time=response_proto.creation_time,
            scorer_id=response_proto.scorer_id,
        )

    def list_scorers(self, experiment_id: str) -> list[ScorerVersion]:
        """
        List all scorers for an experiment (latest version for each scorer name).

        Args:
            experiment_id: String ID of the experiment.

        Returns:
            List of Scorer entities.
        """
        req_body = message_to_json(ListScorers(experiment_id=experiment_id))
        # Scorer APIs are v3.0 endpoints
        response_proto = self._call_endpoint(
            ListScorers,
            req_body,
            endpoint="/api/3.0/mlflow/scorers/list",
        )
        return [ScorerVersion.from_proto(scorer) for scorer in response_proto.scorers]

    def list_scorer_versions(self, experiment_id: str, name: str) -> list[ScorerVersion]:
        """
        List all versions of a specific scorer for an experiment.

        Args:
            experiment_id: String ID of the experiment.
            name: String name of the scorer.

        Returns:
            List of Scorer entities for all versions.
        """
        req_body = message_to_json(ListScorerVersions(experiment_id=experiment_id, name=name))
        # Scorer APIs are v3.0 endpoints
        response_proto = self._call_endpoint(
            ListScorerVersions,
            req_body,
            endpoint="/api/3.0/mlflow/scorers/versions",
        )
        return [ScorerVersion.from_proto(scorer) for scorer in response_proto.scorers]

    def get_scorer(
        self, experiment_id: str, name: str, version: int | None = None
    ) -> ScorerVersion:
        """
        Get a specific scorer for an experiment.

        Args:
            experiment_id: String ID of the experiment.
            name: String name of the scorer.
            version: Integer version of the scorer. If None, returns the scorer
               with maximum version.

        Returns:
            A ScorerVersion entity object.
        """
        req_body = message_to_json(
            GetScorer(experiment_id=experiment_id, name=name, version=version)
        )
        # Scorer APIs are v3.0 endpoints
        response_proto = self._call_endpoint(
            GetScorer,
            req_body,
            endpoint="/api/3.0/mlflow/scorers/get",
        )
        return ScorerVersion.from_proto(response_proto.scorer)

    def delete_scorer(self, experiment_id: str, name: str, version: int | None = None) -> None:
        """
        Delete a scorer for an experiment.

        Args:
            experiment_id: String ID of the experiment.
            name: String name of the scorer.
            version: Integer version of the scorer to delete. If None, deletes all versions.

        Returns:
            None.
        """
        req_body = message_to_json(
            DeleteScorer(experiment_id=experiment_id, name=name, version=version)
        )
        # Scorer APIs are v3.0 endpoints
        self._call_endpoint(
            DeleteScorer,
            req_body,
            endpoint="/api/3.0/mlflow/scorers/delete",
        )

    def upsert_online_scoring_config(
        self,
        experiment_id: str,
        scorer_name: str,
        sample_rate: float,
        filter_string: str | None = None,
    ) -> "OnlineScoringConfig":
        """
        Create or update the online scoring configuration for a registered scorer.

        Args:
            experiment_id: The ID of the Experiment containing the scorer.
            scorer_name: The scorer name.
            sample_rate: The sampling rate (0.0 to 1.0).
            filter_string: Optional filter expression for trace selection.

        Returns:
            The created or updated OnlineScoringConfig object.
        """
        endpoint = "/api/3.0/mlflow/scorers/online-config"
        request_body = {
            "experiment_id": experiment_id,
            "name": scorer_name,
            "sample_rate": sample_rate,
        }
        if filter_string is not None:
            request_body["filter_string"] = filter_string

        response = http_request(
            host_creds=self.get_host_creds(),
            endpoint=endpoint,
            method="PUT",
            json=request_body,
        )

        verify_rest_response(response, endpoint)
        return self._parse_online_scoring_config_from_response(response, endpoint)

    def get_online_scoring_configs(self, scorer_ids: list[str]) -> list["OnlineScoringConfig"]:
        """
        Get online scoring configurations for multiple scorers by their IDs.

        A single scorer can have multiple configurations (e.g., running in different
        experiments or with different filter strings).

        Args:
            scorer_ids: List of scorer IDs to fetch configurations for.

        Returns:
            A list of OnlineScoringConfig objects for the specified scorers.
            Scorers without configurations are not included.
        """
        # Import locally to avoid circular import of RestStore
        from mlflow.genai.scorers.online.entities import OnlineScoringConfig

        if not scorer_ids:
            return []

        endpoint = "/api/3.0/mlflow/scorers/online-configs"
        response = http_request(
            host_creds=self.get_host_creds(),
            endpoint=endpoint,
            method="GET",
            params=[("scorer_ids", sid) for sid in scorer_ids],
        )

        verify_rest_response(response, endpoint)
        try:
            configs_list = response.json()["configs"]
            return [
                OnlineScoringConfig(
                    online_scoring_config_id=config["online_scoring_config_id"],
                    scorer_id=config["scorer_id"],
                    sample_rate=config["sample_rate"],
                    filter_string=config.get("filter_string"),
                    experiment_id=config["experiment_id"],
                )
                for config in configs_list
            ]
        except (KeyError, TypeError, ValueError) as e:
            raise MlflowException(
                f"Unexpected malformed response from {endpoint}: {e}",
                error_code=INTERNAL_ERROR,
            ) from e

    def _parse_online_scoring_config_from_response(self, response, endpoint: str):
        """
        Parse an OnlineScoringConfig from an HTTP response.

        Args:
            response: The HTTP response object.
            endpoint: The API endpoint for error reporting.

        Returns:
            An OnlineScoringConfig instance.

        Raises:
            MlflowException: If the response is malformed.
        """
        # Import locally to avoid circular import of RestStore
        from mlflow.genai.scorers.online.entities import OnlineScoringConfig

        try:
            config_dict = response.json()["config"]
            return OnlineScoringConfig(
                online_scoring_config_id=config_dict["online_scoring_config_id"],
                scorer_id=config_dict["scorer_id"],
                sample_rate=config_dict["sample_rate"],
                filter_string=config_dict.get("filter_string"),
                experiment_id=config_dict["experiment_id"],
            )
        except (KeyError, TypeError, ValueError) as e:
            raise MlflowException(
                f"Unexpected malformed response from {endpoint}: {e}",
                error_code=INTERNAL_ERROR,
            ) from e

    ############################################################################################
    # Deprecated MLflow Tracing APIs. Kept for backward compatibility but do not use.
    ############################################################################################
    def deprecated_start_trace_v2(
        self,
        experiment_id: str,
        timestamp_ms: int,
        request_metadata: dict[str, str],
        tags: dict[str, str],
    ) -> TraceInfoV2:
        """
        DEPRECATED. DO NOT USE.

        Start an initial TraceInfo object in the backend store.

        Args:
            experiment_id: String id of the experiment for this run.
            timestamp_ms: Start time of the trace, in milliseconds since the UNIX epoch.
            request_metadata: Metadata of the trace.
            tags: Tags of the trace.

        Returns:
            The created TraceInfo object.
        """
        request_metadata_proto = []
        for key, value in request_metadata.items():
            attr = TraceRequestMetadata()
            attr.key = key
            attr.value = str(value)
            request_metadata_proto.append(attr)

        tags_proto = []
        for key, value in tags.items():
            tag = TraceTag()
            tag.key = key
            tag.value = str(value)
            tags_proto.append(tag)

        req_body = message_to_json(
            StartTrace(
                experiment_id=str(experiment_id),
                timestamp_ms=timestamp_ms,
                request_metadata=request_metadata_proto,
                tags=tags_proto,
            )
        )
        response_proto = self._call_endpoint(StartTrace, req_body)
        return TraceInfoV2.from_proto(response_proto.trace_info)

    def deprecated_end_trace_v2(
        self,
        request_id: str,
        timestamp_ms: int,
        status: TraceStatus,
        request_metadata: dict[str, str],
        tags: dict[str, str],
    ) -> TraceInfoV2:
        """
        DEPRECATED. DO NOT USE.

        Update the TraceInfo object in the backend store with the completed trace info.

        Args:
            request_id: Unique string identifier of the trace.
            timestamp_ms: End time of the trace, in milliseconds. The execution time field
                in the TraceInfo will be calculated by subtracting the start time from this.
            status: Status of the trace.
            request_metadata: Metadata of the trace. This will be merged with the existing
                metadata logged during the start_trace call.
            tags: Tags of the trace. This will be merged with the existing tags logged
                during the start_trace or set_trace_tag calls.

        Returns:
            The updated TraceInfo object.
        """
        request_metadata_proto = []
        for key, value in request_metadata.items():
            attr = TraceRequestMetadata()
            attr.key = key
            attr.value = str(value)
            request_metadata_proto.append(attr)

        tags_proto = []
        for key, value in tags.items():
            tag = TraceTag()
            tag.key = key
            tag.value = str(value)
            tags_proto.append(tag)

        req_body = message_to_json(
            EndTrace(
                request_id=request_id,
                timestamp_ms=timestamp_ms,
                status=status.to_proto(),
                request_metadata=request_metadata_proto,
                tags=tags_proto,
            )
        )
        # EndTrace endpoint is a dynamic path built with the request_id
        # Always use v2 endpoint (not v3) for this endpoint to maintain compatibility
        endpoint = f"{_REST_API_PATH_PREFIX}/mlflow/traces/{request_id}"
        response_proto = self._call_endpoint(EndTrace, req_body, endpoint=endpoint)
        return TraceInfoV2.from_proto(response_proto.trace_info)

    @databricks_api_disabled(_DATABRICKS_DATASET_API_NAME, _DATABRICKS_DATASET_ALTERNATIVE)
    def create_dataset(
        self,
        name: str,
        tags: dict[str, str] | None = None,
        experiment_ids: list[str] | None = None,
    ) -> "EvaluationDataset":
        """
        Create an evaluation dataset.

        Args:
            name: The name of the evaluation dataset.
            tags: Optional tags to associate with the dataset.
            experiment_ids: List of experiment IDs to associate with the dataset.

        Returns:
            The created EvaluationDataset.
        """
        from mlflow.entities import EvaluationDataset

        req = CreateDataset(
            name=name,
            experiment_ids=experiment_ids or [],
        )

        if tags:
            req.tags = json.dumps(tags)

        req_body = message_to_json(req)
        # Dataset APIs are v3.0 endpoints
        response_proto = self._call_endpoint(
            CreateDataset, req_body, endpoint="/api/3.0/mlflow/datasets/create"
        )
        return EvaluationDataset.from_proto(response_proto.dataset)

    @databricks_api_disabled(_DATABRICKS_DATASET_API_NAME, _DATABRICKS_DATASET_ALTERNATIVE)
    def get_dataset(self, dataset_id: str) -> "EvaluationDataset":
        """
        Get an evaluation dataset by ID.

        Args:
            dataset_id: The ID of the dataset to retrieve.

        Returns:
            The EvaluationDataset object.
        """
        from mlflow.entities import EvaluationDataset

        # GetDataset uses path parameter, not request body
        response_proto = self._call_endpoint(
            GetDataset, None, endpoint=f"/api/3.0/mlflow/datasets/{dataset_id}"
        )
        return EvaluationDataset.from_proto(response_proto.dataset)

    @databricks_api_disabled(_DATABRICKS_DATASET_API_NAME, _DATABRICKS_DATASET_ALTERNATIVE)
    def delete_dataset(self, dataset_id: str) -> None:
        """
        Delete an evaluation dataset.

        Args:
            dataset_id: The ID of the dataset to delete.
        """
        # DeleteDataset uses path parameter, not request body
        self._call_endpoint(DeleteDataset, None, endpoint=f"/api/3.0/mlflow/datasets/{dataset_id}")

    def search_datasets(
        self,
        experiment_ids: list[str] | None = None,
        filter_string: str | None = None,
        max_results: int = 1000,
        order_by: list[str] | None = None,
        page_token: str | None = None,
    ) -> PagedList["EvaluationDataset"]:
        """
        Search for evaluation datasets.

        Args:
            experiment_ids: List of experiment IDs to filter by.
            filter_string: Filter string for dataset names.
            max_results: Maximum number of results to return.
            order_by: Ordering criteria.
            page_token: Token for retrieving the next page of results.

        Returns:
            A PagedList of evaluation datasets.
        """
        from mlflow.entities import EvaluationDataset

        req = SearchEvaluationDatasets(
            experiment_ids=experiment_ids or [],
            filter_string=filter_string,
            max_results=max_results,
            order_by=order_by or [],
            page_token=page_token,
        )
        req_body = message_to_json(req)
        response_proto = self._call_endpoint(
            SearchEvaluationDatasets, req_body, endpoint="/api/3.0/mlflow/datasets/search"
        )
        datasets = [EvaluationDataset.from_proto(ds) for ds in response_proto.datasets]
        return PagedList(datasets, response_proto.next_page_token)

    @databricks_api_disabled(_DATABRICKS_DATASET_API_NAME, _DATABRICKS_DATASET_ALTERNATIVE)
    def upsert_dataset_records(
        self, dataset_id: str, records: list[dict[str, Any]]
    ) -> dict[str, int]:
        """
        Upsert evaluation dataset records.

        Args:
            dataset_id: The ID of the dataset.
            records: List of record dictionaries to upsert.

        Returns:
            Dictionary with 'inserted' and 'updated' counts.
        """
        req = UpsertDatasetRecords(
            records=json.dumps(records),
        )
        req_body = message_to_json(req)
        response_proto = self._call_endpoint(
            UpsertDatasetRecords,
            req_body,
            endpoint=f"/api/3.0/mlflow/datasets/{dataset_id}/records",
        )
        return {
            "inserted": response_proto.inserted_count,
            "updated": response_proto.updated_count,
        }

    @databricks_api_disabled(_DATABRICKS_DATASET_API_NAME, _DATABRICKS_DATASET_ALTERNATIVE)
    def set_dataset_tags(self, dataset_id: str, tags: dict[str, Any]) -> None:
        """
        Set tags for an evaluation dataset.

        This implements an upsert operation - existing tags are merged with new tags.

        Args:
            dataset_id: The ID of the dataset to update.
            tags: Dictionary of tags to update.
        """
        req = SetDatasetTags(
            tags=json.dumps(tags),
        )
        req_body = message_to_json(req)
        self._call_endpoint(
            SetDatasetTags, req_body, endpoint=f"/api/3.0/mlflow/datasets/{dataset_id}/tags"
        )

    @databricks_api_disabled(_DATABRICKS_DATASET_API_NAME, _DATABRICKS_DATASET_ALTERNATIVE)
    def delete_dataset_tag(self, dataset_id: str, key: str) -> None:
        """
        Delete a tag from an evaluation dataset.

        Args:
            dataset_id: The ID of the dataset.
            key: The tag key to delete.
        """
        self._call_endpoint(
            DeleteDatasetTag, None, endpoint=f"/api/3.0/mlflow/datasets/{dataset_id}/tags/{key}"
        )

    @databricks_api_disabled(_DATABRICKS_DATASET_API_NAME, _DATABRICKS_DATASET_ALTERNATIVE)
    def get_dataset_experiment_ids(self, dataset_id: str) -> list[str]:
        """
        Get experiment IDs associated with an evaluation dataset.

        Args:
            dataset_id: The ID of the dataset.

        Returns:
            List of experiment IDs associated with the dataset.
        """
        response_proto = self._call_endpoint(
            GetDatasetExperimentIds,
            None,
            endpoint=f"/api/3.0/mlflow/datasets/{dataset_id}/experiment-ids",
        )
        return list(response_proto.experiment_ids)

    def _load_dataset_records(
        self, dataset_id: str, max_results: int | None = None, page_token: str | None = None
    ) -> tuple["list[DatasetRecord]", str | None]:
        """
        Load dataset records with pagination support.

        Args:
            dataset_id: The ID of the dataset.
            max_results: Maximum number of records to return. If None, returns all records.
            page_token: Token for pagination. If None, starts from the beginning.

        Returns:
            Tuple of (list of DatasetRecord objects, next_page_token).
            next_page_token is None if there are no more records.
        """
        from mlflow.entities.dataset_record import DatasetRecord

        if max_results is None:
            # No pagination requested - fetch all records
            all_records = []
            current_page_token = page_token

            while True:
                req = GetDatasetRecords(max_results=1000)
                if current_page_token:
                    req.page_token = current_page_token

                req_body = message_to_json(req)
                response_proto = self._call_endpoint(
                    GetDatasetRecords,
                    req_body,
                    endpoint=f"/api/3.0/mlflow/datasets/{dataset_id}/records",
                )

                if response_proto.records:
                    records_dicts = json.loads(response_proto.records)
                    all_records.extend(
                        DatasetRecord.from_dict(record_dict) for record_dict in records_dicts
                    )

                if response_proto.next_page_token:
                    current_page_token = response_proto.next_page_token
                else:
                    break

            return all_records, None
        else:
            # Paginated request - fetch only requested page
            req = GetDatasetRecords(max_results=max_results)
            if page_token:
                req.page_token = page_token

            req_body = message_to_json(req)
            response_proto = self._call_endpoint(
                GetDatasetRecords,
                req_body,
                endpoint=f"/api/3.0/mlflow/datasets/{dataset_id}/records",
            )

            records = []
            if response_proto.records:
                records_dicts = json.loads(response_proto.records)
                records.extend(
                    DatasetRecord.from_dict(record_dict) for record_dict in records_dicts
                )

            next_page_token = response_proto.next_page_token or None
            return records, next_page_token

    def link_traces_to_run(self, trace_ids: list[str], run_id: str) -> None:
        """
        Link multiple traces to a run by creating entity associations.

        Args:
            trace_ids: List of trace IDs to link to the run. Maximum 100 traces allowed.
            run_id: ID of the run to link traces to.

        Raises:
            MlflowException: If more than 100 traces are provided.
        """
        req_body = message_to_json(
            LinkTracesToRun(
                trace_ids=trace_ids,
                run_id=run_id,
            )
        )
        self._call_endpoint(LinkTracesToRun, req_body)

    def link_prompts_to_trace(self, trace_id: str, prompt_versions: list[PromptVersion]) -> None:
        """
        Link multiple prompt versions to a trace by creating entity associations.

        Args:
            trace_id: ID of the trace to link prompt versions to.
            prompt_versions: List of PromptVersion objects to link.
        """
        # Convert PromptVersion objects to PromptVersionRef proto messages
        prompt_refs = [
            LinkPromptsToTrace.PromptVersionRef(name=pv.name, version=str(pv.version))
            for pv in prompt_versions
        ]

        req_body = message_to_json(
            LinkPromptsToTrace(
                trace_id=trace_id,
                prompt_versions=prompt_refs,
            )
        )
        self._call_endpoint(LinkPromptsToTrace, req_body)

    def add_dataset_to_experiments(
        self, dataset_id: str, experiment_ids: list[str]
    ) -> "EvaluationDataset":
        """
        Add a dataset to additional experiments via REST API.
        """
        # NB: Local import to avoid circular dependency:
        # rest_store -> EvaluationDataset -> mlflow.data -> artifact_repository_registry
        # -> dbfs_artifact_repo -> rest_store
        from mlflow.entities.evaluation_dataset import EvaluationDataset

        req_body = message_to_json(
            AddDatasetToExperiments(
                dataset_id=dataset_id,
                experiment_ids=experiment_ids,
            )
        )
        response = self._call_endpoint(
            AddDatasetToExperiments,
            req_body,
            endpoint=f"/api/3.0/mlflow/datasets/{dataset_id}/add-experiments",
        )
        return EvaluationDataset.from_proto(response.dataset)

    def remove_dataset_from_experiments(
        self, dataset_id: str, experiment_ids: list[str]
    ) -> "EvaluationDataset":
        """
        Remove a dataset from experiments via REST API.
        """
        # NB: Local import to avoid circular dependency:
        # rest_store -> EvaluationDataset -> mlflow.data -> artifact_repository_registry
        # -> dbfs_artifact_repo -> rest_store
        from mlflow.entities.evaluation_dataset import EvaluationDataset

        req_body = message_to_json(
            RemoveDatasetFromExperiments(
                dataset_id=dataset_id,
                experiment_ids=experiment_ids,
            )
        )
        response = self._call_endpoint(
            RemoveDatasetFromExperiments,
            req_body,
            endpoint=f"/api/3.0/mlflow/datasets/{dataset_id}/remove-experiments",
        )
        return EvaluationDataset.from_proto(response.dataset)

    def log_spans(self, location: str, spans: list[Span], tracking_uri=None) -> list[Span]:
        """
        Log multiple span entities to the tracking store via the OTel API.

        Args:
            location: The location to log spans to. It should be experiment ID of an MLflow
                experiment.
            spans: List of Span entities to log. All spans must belong to the same trace.
            tracking_uri: The tracking URI to use. Default to None.

        Returns:
            List of logged Span entities.

        Raises:
            MlflowException: If spans belong to different traces or the OTel API call fails.
        """
        if not spans:
            return []

        server_version = self._get_server_version(self.get_host_creds())
        if server_version is None:
            raise NotImplementedError(
                "log_spans is not supported: could not identify MLflow server version"
            )
        elif server_version < Version("3.4"):
            raise NotImplementedError(
                f"log_spans is not supported: MLflow server version {server_version} is"
                f" less than 3.4"
            )

        trace_ids = {span.trace_id for span in spans}
        if len(trace_ids) > 1:
            raise MlflowException(
                f"All spans must belong to the same trace. Found trace IDs: {trace_ids}",
                error_code=databricks_pb2.INVALID_PARAMETER_VALUE,
            )

        request = ExportTraceServiceRequest()
        resource_spans = request.resource_spans.add()
        scope_spans = resource_spans.scope_spans.add()
        scope_spans.spans.extend(span.to_otel_proto() for span in spans)

        response = http_request(
            host_creds=self.get_host_creds(),
            endpoint=OTLP_TRACES_PATH,
            method="POST",
            data=request.SerializeToString(),
            extra_headers={
                "Content-Type": "application/x-protobuf",
                MLFLOW_EXPERIMENT_ID_HEADER: location,
            },
        )

        verify_rest_response(response, OTLP_TRACES_PATH)
        return spans

    async def log_spans_async(self, location: str, spans: list[Span]) -> list[Span]:
        """
        Async wrapper for log_spans method.

        Args:
            location: The location to log spans to. It should be experiment ID of an MLflow
                experiment.
            spans: List of Span entities to log. All spans must belong to the same trace.

        Returns:
            List of logged Span entities.

        Raises:
            MlflowException: If spans belong to different traces or the OTel API call fails.
        """
        return self.log_spans(location, spans)



"""
Assistant API endpoints for MLflow Server.

This module provides endpoints for integrating AI assistants with MLflow UI,
enabling AI-powered helper through a chat interface.
"""

import ipaddress
import uuid
from pathlib import Path
from typing import Any, AsyncGenerator

from fastapi import APIRouter, Depends, HTTPException, Request
from fastapi.responses import StreamingResponse
from pydantic import BaseModel, Field

from mlflow.assistant import get_project_path
from mlflow.assistant.providers.claude_code import ClaudeCodeProvider
from mlflow.assistant.types import EventType
from mlflow.server.assistant.session import SessionManager

# TODO: Hardcoded provider until supporting multiple providers
_provider = ClaudeCodeProvider()


# Update the message when we support proxy access
_BLOCK_REMOTE_ACCESS_ERROR_MSG = (
    "Assistant API is only accessible from the same host where the mLflow server is running."
)


async def _require_localhost(request: Request) -> None:
    """
    Dependency that restricts access to localhost only.

    Uses ipaddress library for robust loopback detection.

    Raises:
        HTTPException: If request is not from localhost
    """
    client_host = request.client.host if request.client else None

    if not client_host:
        raise HTTPException(status_code=403, detail=_BLOCK_REMOTE_ACCESS_ERROR_MSG)

    try:
        ip = ipaddress.ip_address(client_host)
    except ValueError:
        raise HTTPException(status_code=403, detail=_BLOCK_REMOTE_ACCESS_ERROR_MSG)

    if not ip.is_loopback:
        raise HTTPException(status_code=403, detail=_BLOCK_REMOTE_ACCESS_ERROR_MSG)


assistant_router = APIRouter(
    prefix="/ajax-api/3.0/mlflow/assistant",
    tags=["assistant"],
    dependencies=[Depends(_require_localhost)],
)


class MessageRequest(BaseModel):
    message: str
    session_id: str | None = None  # empty for the first message
    experiment_id: str | None = None
    context: dict[str, Any] = Field(default_factory=dict)


class MessageResponse(BaseModel):
    session_id: str
    stream_url: str


class GetStatusResponse(BaseModel):
    provider: str
    available: bool


@assistant_router.post("/message")
async def send_message(request: MessageRequest) -> MessageResponse:
    """
    Send a message to the assistant and get a session for streaming the response.

    Args:
        request: MessageRequest with message, context, and optional session_id

    Returns:
        MessageResponse with session_id and stream_url
    """
    # Generate or use existing session ID
    session_id = request.session_id or str(uuid.uuid4())

    project_path = get_project_path(request.experiment_id) if request.experiment_id else None

    # Create or update session
    session = SessionManager.load(session_id)
    if session is None:
        session = SessionManager.create(
            context=request.context, working_dir=Path(project_path) if project_path else None
        )
    elif request.context:
        session.update_context(request.context)

    # Store the pending message with role
    session.set_pending_message(role="user", content=request.message)
    session.add_message(role="user", content=request.message)
    SessionManager.save(session_id, session)

    return MessageResponse(
        session_id=session_id,
        stream_url=f"/ajax-api/3.0/mlflow/assistant/stream/{session_id}",
    )


@assistant_router.get("/sessions/{session_id}/stream")
async def stream_response(session_id: str) -> StreamingResponse:
    """
    Stream the assistant's response via Server-Sent Events.

    Args:
        session_id: The session ID returned from /message

    Returns:
        StreamingResponse with SSE events
    """
    session = SessionManager.load(session_id)
    if session is None:
        raise HTTPException(status_code=404, detail="Session not found")

    # Get and clear the pending message
    pending_message = session.clear_pending_message()
    if not pending_message:
        raise HTTPException(status_code=400, detail="No pending message to process")
    SessionManager.save(session_id, session)

    async def event_generator() -> AsyncGenerator[str, None]:
        nonlocal session
        async for event in _provider.astream(
            prompt=pending_message.content,
            session_id=session.provider_session_id,
            cwd=session.working_dir,
            context=session.context,
        ):
            # Store provider session ID if returned (for conversation continuity)
            if event.type == EventType.DONE:
                session.provider_session_id = event.data.get("session_id")
                SessionManager.save(session_id, session)

            yield event.to_sse_event()

    return StreamingResponse(
        event_generator(),
        media_type="text/event-stream",
        headers={
            "Cache-Control": "no-cache",
            "Connection": "keep-alive",
            "X-Accel-Buffering": "no",
        },
    )


@assistant_router.get("/status")
async def get_status() -> GetStatusResponse:
    """
    Get the status of the assistant provider.

    Returns:
        Provider status including availability
    """
    return GetStatusResponse(
        provider=_provider.name,
        available=_provider.is_available(),
    )



"""add trace metrics table

Create Date: 2025-12-04 12:00:00.000000

"""

import sqlalchemy as sa
from alembic import op

from mlflow.store.tracking.dbmodels.models import SqlTraceMetrics

# revision identifiers, used by Alembic.
revision = "b7c8d9e0f1a2"
down_revision = "1bd49d398cd23"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        SqlTraceMetrics.__tablename__,
        sa.Column("request_id", sa.String(length=50), nullable=False),
        sa.Column("key", sa.String(length=250), nullable=False),
        sa.Column("value", sa.Float(precision=53), nullable=True),
        sa.ForeignKeyConstraint(
            ["request_id"],
            ["trace_info.request_id"],
            name="fk_trace_metrics_request_id",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("request_id", "key", name="trace_metrics_pk"),
    )

    # Add index on request_id for faster lookups
    with op.batch_alter_table(SqlTraceMetrics.__tablename__, schema=None) as batch_op:
        batch_op.create_index(
            f"index_{SqlTraceMetrics.__tablename__}_request_id",
            ["request_id"],
            unique=False,
        )


def downgrade():
    op.drop_table(SqlTraceMetrics.__tablename__)



import logging
from typing import TYPE_CHECKING, Any

from mlflow.entities.assessment import Assessment, AssessmentSource, Feedback
from mlflow.entities.assessment_source import AssessmentSourceType
from mlflow.entities.trace import Trace
from mlflow.exceptions import MlflowException
from mlflow.genai.judges.base import AlignmentOptimizer, Judge, JudgeField
from mlflow.genai.judges.optimizers.dspy_utils import (
    _check_dspy_installed,
    construct_dspy_lm,
    create_dspy_signature,
    trace_to_dspy_example,
)
from mlflow.genai.judges.optimizers.memalign.utils import (
    Guideline,
    create_extended_signature,
    distill_guidelines,
    get_default_embedding_model,
    retrieve_relevant_examples,
    truncate_to_token_limit,
)
from mlflow.genai.judges.utils import get_default_model
from mlflow.protos.databricks_pb2 import INTERNAL_ERROR, INVALID_PARAMETER_VALUE
from mlflow.utils.annotations import experimental
from mlflow.utils.docstring_utils import format_docstring

if TYPE_CHECKING:
    import dspy

_logger = logging.getLogger(__name__)

_MODEL_API_DOC = {
    "reflection_lm": """Model to use for distilling guidelines from feedback.
Supported formats:

* `"databricks"` for Databricks-native integration
* `"databricks:/<endpoint-name>"` or `"endpoints:/<endpoint-name>"` for
  Databricks model serving endpoints
* `<provider>:/<model-name>` for other providers (e.g.,
  `"openai:/gpt-4o-mini"`, `"anthropic:/claude-3.5-sonnet-20240620"`)

MLflow natively supports `["openai", "anthropic", "bedrock", "mistral"]`,
and more providers are supported through
`LiteLLM <https://docs.litellm.ai/docs/providers>`_.

Default model depends on the tracking URI setup:

* Databricks: `databricks`
* Otherwise: `openai:/gpt-4o-mini`.
""",
    "embedding_model": """Model to use for generating embeddings for
example retrieval. Must be a form of `<provider>/<model-name>`, such as
`"openai/text-embedding-3-small"`. Supported providers include OpenAI and
others via LiteLLM. Default: `"openai/text-embedding-3-small"`.
""",
}


@experimental(version="3.9.0")
@format_docstring(_MODEL_API_DOC)
class MemoryAugmentedJudge(Judge):
    """
    A judge augmented with dual memory systems.

    This judge enhances evaluation with:
    - Semantic Memory: Distilled guidelines from past feedback
    - Episodic Memory: Retrieved similar examples from past feedback

    The judge maintains state across evaluations and provides the exact same
    interface as the original judge. Memories are managed internally.

    Args:
        base_judge: Base judge to augment with memory systems
        reflection_lm: {{ reflection_lm }}
        retrieval_k: Number of similar examples to retrieve from episodic memory (default: 5)
        embedding_model: {{ embedding_model }}
        embedding_dim: Dimension of embeddings (default: 512)
        episodic_memory: Initial examples to add to episodic memory
        semantic_memory: Initial guidelines to add to semantic memory. If None and base_judge
            is a MemoryAugmentedJudge, inherits semantic memory from base_judge.
    """

    def __init__(
        self,
        base_judge: Judge,
        reflection_lm: str | None = None,
        retrieval_k: int = 5,
        embedding_model: str | None = None,
        embedding_dim: int = 512,
        episodic_memory: list["dspy.Example"] | None = None,
        semantic_memory: list[Guideline] | None = None,
    ):
        import dspy

        effective_base_judge = (
            base_judge._base_judge if isinstance(base_judge, MemoryAugmentedJudge) else base_judge
        )

        # Determine if we should skip distillation during initialization
        # Skip if user explicitly provides semantic_memory (they want to use those exact guidelines)
        skip_initial_distillation = semantic_memory is not None

        if semantic_memory is not None:
            initial_guidelines = list(semantic_memory)
        elif isinstance(base_judge, MemoryAugmentedJudge):
            initial_guidelines = list(base_judge._semantic_memory)
        else:
            initial_guidelines = []

        super().__init__(
            name=effective_base_judge.name,
            description=effective_base_judge.description,
            aggregations=effective_base_judge.aggregations,
        )

        self._base_judge = effective_base_judge
        self._base_signature = create_dspy_signature(effective_base_judge)
        self._retrieval_k = retrieval_k
        self._episodic_memory: list["dspy.Example"] = []
        self._semantic_memory: list[Guideline] = initial_guidelines

        self._reflection_lm = reflection_lm if reflection_lm is not None else get_default_model()

        self._embedding_model = (
            embedding_model if embedding_model is not None else get_default_embedding_model()
        )

        # TODO: Add support for Databricks embedding models with DSPy embedder
        if self._embedding_model.startswith("databricks:/"):
            raise MlflowException(
                "Databricks embedding models are not currently supported for MemAlign. "
                f"Please use a different embedding model (e.g., 'openai/text-embedding-3-small'). "
                f"Got: {self._embedding_model}",
                error_code=INVALID_PARAMETER_VALUE,
            )

        self._embedding_dim = embedding_dim
        self._embedder = dspy.Embedder(self._embedding_model, dimensions=self._embedding_dim)
        self._retriever = None

        extended_signature = create_extended_signature(self._base_signature)
        self._predict_module = dspy.Predict(extended_signature)
        self._predict_module.set_lm(construct_dspy_lm(effective_base_judge.model))

        if episodic_memory:
            if skip_initial_distillation:
                # User provided explicit guidelines, just add examples without distilling
                self._episodic_memory.extend(episodic_memory)
                self._build_episodic_memory()
            else:
                # Normal flow: add examples and distill guidelines
                self._add_examples_to_memory(episodic_memory)

    def __call__(
        self,
        *,
        inputs: Any = None,
        outputs: Any = None,
        expectations: dict[str, Any] | None = None,
        trace: Trace | None = None,
    ) -> Assessment:
        guidelines = [g.guideline_text for g in self._semantic_memory]
        query_kwargs = {
            "inputs": inputs,
            "outputs": outputs,
            "expectations": expectations,
            "trace": trace,
        }
        retrieved_results = retrieve_relevant_examples(
            retriever=self._retriever,
            examples=self._episodic_memory,
            query_kwargs=query_kwargs,
            signature=self._base_signature,
        )

        relevant_examples = [example for example, _ in retrieved_results]
        retrieved_trace_ids = [trace_id for _, trace_id in retrieved_results]

        prediction = self._predict_module(
            guidelines=guidelines,
            example_judgements=relevant_examples,
            inputs=inputs,
            outputs=outputs,
            expectations=expectations,
            trace=trace,
        )

        return Feedback(
            name=self._base_judge.name,
            source=AssessmentSource(
                source_type=AssessmentSourceType.LLM_JUDGE,
                source_id=self._base_judge.model,
            ),
            value=prediction.result,
            rationale=prediction.rationale,
            metadata={"retrieved_example_trace_ids": retrieved_trace_ids}
            if retrieved_trace_ids
            else {},
        )

    @property
    def name(self) -> str:
        return self._base_judge.name

    @property
    def instructions(self) -> str:
        instructions = self._base_judge.instructions
        if self._semantic_memory:
            instructions += f"\n\nDistilled Guidelines ({len(self._semantic_memory)}):\n"
            for guideline in self._semantic_memory:
                instructions += f"  - {guideline.guideline_text}\n"
        return instructions

    @property
    def model(self) -> str:
        return self._base_judge.model

    def get_input_fields(self) -> list[JudgeField]:
        return self._base_judge.get_input_fields()

    @experimental(version="3.9.0")
    def unalign(self, traces: list[Trace]) -> "MemoryAugmentedJudge":
        """
        Remove specific traces from memory and return new judge.

        This method allows you to selectively remove feedback examples from the judge's
        memory systems. This is useful when you want to:

        - Remove incorrect or low-quality feedback that negatively impacts performance
        - Update the judge in case your evaluation criteria change
        - Remove feedback from specific users or time periods

        The returned judge will have guidelines selectively deleted based on source_trace_ids:
        - Guidelines where any source trace was removed are deleted
        - Guidelines with no removed source traces are retained
        - Guidelines without source_trace_ids are always retained

        Args:
            traces: Traces containing feedback to remove from memory. Only traces with
                feedback matching this judge's name will be removed.

        Returns:
            A new MemoryAugmentedJudge with the specified traces removed from memory.

        Example:
            .. code-block:: python

                import mlflow
                from mlflow.genai.judges import make_judge
                from mlflow.genai.judges.optimizers import MemAlignOptimizer

                # Assuming `all_traces` contains human feedback for the judge
                aligned_judge = judge.align(traces=all_traces, optimizer=MemAlignOptimizer())
                aligned_judge_v2 = aligned_judge.unalign(traces=bad_traces)
                # The judge now only reflects feedback from `set(all_traces) - set(bad_traces)`
        """
        trace_ids_to_remove = {trace.info.trace_id for trace in traces}

        # Filter examples to remove
        filtered_examples = [
            example
            for example in self._episodic_memory
            if not (hasattr(example, "_trace_id") and example._trace_id in trace_ids_to_remove)
        ]

        if len(filtered_examples) == len(self._episodic_memory):
            _logger.warning("No feedback records found for the provided traces")
            return self

        # Filter guidelines based on source_trace_ids
        # - Always retain user-provided guidelines (those without source_trace_ids)
        # - Delete guideline only if ALL of its source traces were removed
        retained_guidelines = [
            guideline
            for guideline in self._semantic_memory
            if guideline.source_trace_ids is None
            or any(tid not in trace_ids_to_remove for tid in guideline.source_trace_ids)
        ]

        return MemoryAugmentedJudge(
            base_judge=self._base_judge,
            reflection_lm=self._reflection_lm,
            retrieval_k=self._retrieval_k,
            embedding_model=self._embedding_model,
            embedding_dim=self._embedding_dim,
            episodic_memory=filtered_examples,
            semantic_memory=retained_guidelines,
        )

    def _distill_new_guidelines(self, new_examples: list["dspy.Example"]) -> None:
        """
        Distill new guidelines from newly added examples and add to semantic memory.

        Args:
            new_examples: The examples that were just added (not all examples)
        """
        existing_guideline_texts = [g.guideline_text for g in self._semantic_memory]
        new_guidelines = distill_guidelines(
            examples=new_examples,
            signature=self._base_signature,
            judge_instructions=self._base_judge.instructions,
            reflection_lm=self._reflection_lm,
            existing_guidelines=existing_guideline_texts,
        )
        self._semantic_memory.extend(new_guidelines)
        _logger.debug(
            f"Distilled {len(new_guidelines)} new guidelines from {len(new_examples)} new examples"
        )

    def _build_episodic_memory(self) -> None:
        """Build episodic memory search index from examples."""
        import dspy.retrievers

        # Build episodic memory corpus from input fields
        corpus = []
        for example in self._episodic_memory:
            query_parts = []
            for field_name in self._base_signature.input_fields:
                if hasattr(example, field_name):
                    value = getattr(example, field_name)
                    if value is not None:
                        query_parts.append(str(value))
            query = " ".join(query_parts)
            query = truncate_to_token_limit(query, self._embedding_model)
            corpus.append(query)

        self._retriever = dspy.retrievers.Embeddings(
            embedder=self._embedder, corpus=corpus, k=self._retrieval_k
        )
        _logger.debug(f"Episodic memory corpus contains {len(corpus)} examples")

    def _add_examples_to_memory(self, examples: list["dspy.Example"]) -> None:
        """Add examples to episodic memory and distill new guidelines.

        Args:
            examples: Examples to add to episodic memory
        """
        self._episodic_memory.extend(examples)
        self._distill_new_guidelines(examples)
        self._build_episodic_memory()


@experimental(version="3.9.0")
@format_docstring(_MODEL_API_DOC)
class MemAlignOptimizer(AlignmentOptimizer):
    """
    MemAlign alignment optimizer using dual memory systems.

    This optimizer creates a memory-augmented judge that learns from feedback
    through two complementary mechanisms:

    **Semantic Memory** - Distills general guidelines from feedback:
        - LLM extracts patterns from feedback records
        - Guidelines describe user preferences and expectations
        - Applied as context to all future evaluations

    **Episodic Memory** - Retrieves similar past examples:
        - Stores feedback records with embeddings
        - Finds most similar examples during evaluation
        - Provides concrete examples as evaluation context

    The returned judge is a MemoryAugmentedJudge that maintains memory state.

    Args:
        reflection_lm: {{ reflection_lm }}
        retrieval_k: Number of similar examples to retrieve from episodic memory (default: 5)
        embedding_model: {{ embedding_model }}
        embedding_dim: Dimension of embeddings (default: 512)

    Example:
        .. code-block:: python

            import mlflow
            from mlflow.genai.judges import make_judge
            from mlflow.genai.judges.optimizers import MemAlignOptimizer

            judge = make_judge(name="quality", instructions="...", model="openai:/gpt-4")
            optimizer = MemAlignOptimizer(
                reflection_lm="openai:/gpt-4o-mini",
                retrieval_k=3,
                embedding_model="openai/text-embedding-3-small",
            )

            # Assuming `traces` contains human feedback for the judge
            optimized_judge = judge.align(traces=traces, optimizer=optimizer)
            result = optimized_judge(inputs="...", outputs="...")
    """

    def __init__(
        self,
        reflection_lm: str | None = None,
        retrieval_k: int = 5,
        embedding_model: str | None = None,
        embedding_dim: int = 512,
    ):
        _check_dspy_installed()
        self._reflection_lm = reflection_lm if reflection_lm is not None else get_default_model()
        self._retrieval_k = retrieval_k
        self._embedding_model = (
            embedding_model if embedding_model is not None else get_default_embedding_model()
        )
        self._embedding_dim = embedding_dim

    def align(self, judge: Judge, traces: list[Trace]) -> MemoryAugmentedJudge:
        """
        Align judge with human feedback from traces.

        Args:
            judge: Judge to align
            traces: Traces containing human feedback

        Returns:
            Memory-augmented judge aligned with feedback
        """
        try:
            if not traces:
                raise MlflowException(
                    "No traces provided for alignment", error_code=INVALID_PARAMETER_VALUE
                )

            _logger.debug(f"Starting MemAlign alignment with {len(traces)} traces")

            existing_examples = (
                list(judge._episodic_memory) if isinstance(judge, MemoryAugmentedJudge) else []
            )

            new_examples = []
            for trace in traces:
                example = trace_to_dspy_example(trace, judge)
                if example is not None:
                    example._trace_id = trace.info.trace_id
                    new_examples.append(example)

            if not new_examples:
                raise MlflowException(
                    f"No valid feedback records found in traces. "
                    f"Ensure traces contain human assessments with name '{judge.name}'",
                    error_code=INVALID_PARAMETER_VALUE,
                )

            _logger.debug(
                f"Created {len(new_examples)} new feedback records from {len(traces)} traces"
            )

            all_examples = existing_examples + new_examples

            memory_judge = MemoryAugmentedJudge(
                base_judge=judge,
                reflection_lm=self._reflection_lm,
                retrieval_k=self._retrieval_k,
                embedding_model=self._embedding_model,
                embedding_dim=self._embedding_dim,
                episodic_memory=all_examples,
            )

            _logger.debug(
                f"MemAlign alignment completed successfully. Aligned {len(new_examples)} examples."
            )
            return memory_judge

        except Exception as e:
            _logger.error(f"MemAlign alignment failed: {e}", exc_info=True)
            raise MlflowException(
                f"Alignment optimization failed: {e!s}", error_code=INTERNAL_ERROR
            ) from e



from __future__ import annotations

from typing import Any

from mlflow.entities.trace import Trace
from mlflow.exceptions import MlflowException
from mlflow.genai.utils.trace_utils import (
    extract_retrieval_context_from_trace,
    parse_inputs_to_str,
    parse_outputs_to_str,
    resolve_expectations_from_trace,
    resolve_inputs_from_trace,
    resolve_outputs_from_trace,
)


class _NoOpRateLimiter:
    """Minimal rate limiter stub for Phoenix model compatibility."""

    def __init__(self):
        self._verbose = False


def check_phoenix_installed():
    """Check if Phoenix is installed and raise an error if not."""
    try:
        import phoenix.evals  # noqa: F401
    except ImportError:
        raise MlflowException.invalid_parameter_value(
            "Phoenix evaluators require the 'arize-phoenix-evals' package. "
            "Install it with: pip install arize-phoenix-evals"
        )


def map_scorer_inputs_to_phoenix_record(
    inputs: Any = None,
    outputs: Any = None,
    expectations: dict[str, Any] | None = None,
    trace: Trace | None = None,
) -> dict[str, Any]:
    """
    Convert MLflow scorer inputs to Phoenix evaluator record format.

    Args:
        inputs: The input to evaluate
        outputs: The output to evaluate
        expectations: Expected values and context for evaluation
        trace: MLflow trace for evaluation

    Returns:
        Dictionary formatted for Phoenix evaluator
    """
    if trace:
        inputs = resolve_inputs_from_trace(inputs, trace)
        outputs = resolve_outputs_from_trace(outputs, trace)
        expectations = resolve_expectations_from_trace(expectations, trace)

    record = {}

    if inputs is not None:
        record["input"] = parse_inputs_to_str(inputs)

    if outputs is not None:
        record["output"] = parse_outputs_to_str(outputs)

    # Extract reference from expectations using standard MLflow keys
    reference = None
    if expectations:
        reference = (
            expectations.get("expected_response")
            or expectations.get("context")
            or expectations.get("reference")
            or expectations.get("expected_output")
        )

    # Fall back to extracting context from trace retrieval spans
    if not reference and trace:
        if span_id_to_context := extract_retrieval_context_from_trace(trace):
            contexts = []
            for ctx_list in span_id_to_context.values():
                for ctx in ctx_list:
                    if isinstance(ctx, dict) and "content" in ctx:
                        contexts.append(ctx["content"])
                    else:
                        contexts.append(str(ctx))
            reference = "\n".join(contexts) if contexts else None

    if reference:
        record["reference"] = reference

    return record



import json
import os
import tempfile
import uuid
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

from mlflow.assistant.types import Message

SESSION_DIR = Path(tempfile.gettempdir()) / "mlflow-assistant-sessions"


@dataclass
class Session:
    """Session state for assistant conversations."""

    context: dict[str, Any] = field(default_factory=dict)
    messages: list[Message] = field(default_factory=list)
    pending_message: Message | None = None
    provider_session_id: str | None = None
    working_dir: Path | None = None  # Working directory for the session (e.g. project path)

    def add_message(self, role: str, content: str) -> None:
        """Add a message to the session history.

        Args:
            role: Role of the message sender (user, assistant, system)
            content: Text content of the message
        """
        self.messages.append(Message(role=role, content=content))

    def set_pending_message(self, role: str, content: str) -> None:
        """Set the pending message to be processed.

        Args:
            role: Role of the message sender
            content: Text content of the message
        """
        self.pending_message = Message(role=role, content=content)

    def clear_pending_message(self) -> Message | None:
        """Clear and return the pending message.

        Returns:
            The pending message, or None if no message was pending
        """
        msg = self.pending_message
        self.pending_message = None
        return msg

    def update_context(self, context: dict[str, Any]) -> None:
        """Update session context.

        Args:
            context: Context data to merge into session context
        """
        self.context.update(context)

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary for serialization.

        Returns:
            Dictionary representation of session
        """
        return {
            "context": self.context,
            "messages": [msg.model_dump() for msg in self.messages],
            "pending_message": self.pending_message.model_dump() if self.pending_message else None,
            "provider_session_id": self.provider_session_id,
            "working_dir": self.working_dir.as_posix() if self.working_dir else None,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "Session":
        """Load from dictionary.

        Args:
            data: Dictionary representation of session

        Returns:
            Session instance
        """
        messages = [Message.model_validate(m) for m in data.get("messages", [])]
        pending = data.get("pending_message")
        pending_msg = Message.model_validate(pending) if pending else None

        return cls(
            context=data.get("context", {}),
            messages=messages,
            pending_message=pending_msg,
            provider_session_id=data.get("provider_session_id"),
            working_dir=Path(data.get("working_dir")) if data.get("working_dir") else None,
        )


class SessionManager:
    """Manages session storage and retrieval.

    Provides static methods for session operations, keeping
    Session as a simple data container.
    """

    @staticmethod
    def validate_session_id(session_id: str) -> None:
        """Validate that session_id is a valid UUID to prevent path traversal.

        Args:
            session_id: Session ID to validate

        Raises:
            ValueError: If session ID is not a valid UUID
        """
        try:
            uuid.UUID(session_id)
        except (ValueError, TypeError) as e:
            raise ValueError("Invalid session ID format") from e

    @staticmethod
    def get_session_file(session_id: str) -> Path:
        """Get the file path for a session.

        Args:
            session_id: Session ID

        Returns:
            Path to session file

        Raises:
            ValueError: If session ID is invalid
        """
        SessionManager.validate_session_id(session_id)
        return SESSION_DIR / f"{session_id}.json"

    @staticmethod
    def save(session_id: str, session: Session) -> None:
        """Save session to disk atomically.

        Args:
            session_id: Session ID
            session: Session to save

        Raises:
            ValueError: If session ID is invalid
        """
        SessionManager.validate_session_id(session_id)
        SESSION_DIR.mkdir(parents=True, exist_ok=True)
        session_file = SessionManager.get_session_file(session_id)

        # Write to temp file, then rename (atomic on POSIX)
        fd, temp_path = tempfile.mkstemp(dir=SESSION_DIR, suffix=".tmp")
        try:
            with os.fdopen(fd, "w") as f:
                json.dump(session.to_dict(), f)
            os.replace(temp_path, session_file)
        except Exception:
            os.unlink(temp_path)
            raise

    @staticmethod
    def load(session_id: str) -> Session | None:
        """Load session from disk.

        Args:
            session_id: Session ID

        Returns:
            Session instance, or None if not found
        """
        try:
            session_file = SessionManager.get_session_file(session_id)
        except ValueError:
            return None
        if not session_file.exists():
            return None
        data = json.loads(session_file.read_text())
        return Session.from_dict(data)

    @staticmethod
    def create(context: dict[str, Any] | None = None, working_dir: Path | None = None) -> Session:
        """Create a new session.

        Args:
            context: Initial context data, or None
            working_dir: Working directory for the session

        Returns:
            New Session instance
        """
        return Session(context=context or {}, working_dir=working_dir)



import json
from enum import Enum
from typing import Any, Literal

from pydantic import BaseModel, Field

# Message interface between assistant providers and the assistant client
# Inspired by https://github.com/anthropics/claude-agent-sdk-python/blob/29c12cd80b256e88f321b2b8f1f5a88445077aa5/src/claude_agent_sdk/types.py


class TextBlock(BaseModel):
    """Text content block."""

    text: str


class ThinkingBlock(BaseModel):
    """Thinking content block."""

    thinking: str
    signature: str


class ToolUseBlock(BaseModel):
    """Tool use content block."""

    id: str
    name: str
    input: dict[str, Any]


class ToolResultBlock(BaseModel):
    """Tool result content block."""

    tool_use_id: str
    content: str | list[dict[str, Any]] | None = None
    is_error: bool | None = None


ContentBlock = TextBlock | ThinkingBlock | ToolUseBlock | ToolResultBlock


class Message(BaseModel):
    """Structured message representation for assistant conversations.

    Uses standard chat message format with role and content fields.
    Can be extended in the future to support multi-modal content.
    """

    role: Literal["user", "assistant", "system"] = Field(description="Role of the message sender")
    content: str | list[ContentBlock] = Field(description="Content of the message")


class EventType(str, Enum):
    MESSAGE = "message"
    STREAM_EVENT = "stream_event"
    DONE = "done"
    ERROR = "error"


class Event(BaseModel):
    """A common event format parsed from the raw assistant provider output."""

    type: EventType
    data: dict[str, Any]

    def to_sse_event(self) -> str:
        """Convert the event to an SSE event string."""
        return f"event: {self.type}\ndata: {json.dumps(self.data)}\n\n"

    @classmethod
    def from_error(cls, error: str) -> "Event":
        return cls(type=EventType.ERROR, data={"error": error})

    @classmethod
    def from_message(cls, message: Message) -> "Event":
        return cls(type=EventType.MESSAGE, data={"message": message.model_dump()})

    @classmethod
    def from_stream_event(cls, event: dict[str, Any]) -> "Event":
        return cls(type=EventType.STREAM_EVENT, data={"event": event})

    @classmethod
    def from_result(cls, result: Any, session_id: str) -> "Event":
        return cls(type=EventType.DONE, data={"result": result, "session_id": session_id})



"""update job table

Create Date: 2025-12-16 16:31:47.921120

"""

from alembic import op
from sqlalchemy import String

# revision identifiers, used by Alembic.
revision = "5d2d30f0abce"
down_revision = "b7c8d9e0f1a2"
branch_labels = None
depends_on = None


def upgrade():
    # Rename column `function_fullname` -> `job_name` and update the related index
    with op.batch_alter_table("jobs", schema=None) as batch_op:
        # Drop old index that referenced `function_fullname`
        batch_op.drop_index("index_jobs_function_status_creation_time")
        # Rename the column
        batch_op.alter_column(
            "function_fullname", new_column_name="job_name", existing_type=String(500)
        )

    with op.batch_alter_table("jobs", schema=None) as batch_op:
        # Recreate the index referencing the new column name
        batch_op.create_index(
            "index_jobs_name_status_creation_time",
            ["job_name", "status", "creation_time"],
            unique=False,
        )


def downgrade():
    # Revert column rename `job_name` -> `function_fullname` and restore the original index
    with op.batch_alter_table("jobs", schema=None) as batch_op:
        batch_op.drop_index("index_jobs_name_status_creation_time")
        batch_op.alter_column(
            "job_name", new_column_name="function_fullname", existing_type=String(500)
        )

    with op.batch_alter_table("jobs", schema=None) as batch_op:
        batch_op.create_index(
            "index_jobs_function_status_creation_time",
            ["function_fullname", "status", "creation_time"],
            unique=False,
        )



"""Checkpoint management for trace-level online scoring."""

import json
import logging
import time
from dataclasses import asdict, dataclass

from mlflow.entities.experiment_tag import ExperimentTag
from mlflow.genai.scorers.online.constants import MAX_LOOKBACK_MS
from mlflow.store.tracking.abstract_store import AbstractStore
from mlflow.utils.mlflow_tags import MLFLOW_LATEST_ONLINE_SCORING_TRACE_CHECKPOINT

_logger = logging.getLogger(__name__)


@dataclass
class OnlineTraceScoringCheckpoint:
    timestamp_ms: int  # Timestamp of the last processed trace in milliseconds
    trace_id: str | None = None  # Trace ID used as tie breaker when traces have same timestamp

    def to_json(self) -> str:
        return json.dumps(asdict(self))

    @classmethod
    def from_json(cls, json_str: str) -> "OnlineTraceScoringCheckpoint":
        data = json.loads(json_str)
        return cls(**data)


@dataclass
class OnlineTraceScoringTimeWindow:
    min_trace_timestamp_ms: int
    max_trace_timestamp_ms: int


class OnlineTraceCheckpointManager:
    def __init__(self, tracking_store: AbstractStore, experiment_id: str):
        self._tracking_store = tracking_store
        self._experiment_id = experiment_id

    def get_checkpoint(self) -> OnlineTraceScoringCheckpoint | None:
        """
        Get the last processed trace checkpoint from the experiment tag.

        Returns:
            OnlineTraceScoringCheckpoint, or None if no checkpoint exists.
        """
        try:
            experiment = self._tracking_store.get_experiment(self._experiment_id)
            if checkpoint_str := experiment.tags.get(MLFLOW_LATEST_ONLINE_SCORING_TRACE_CHECKPOINT):
                return OnlineTraceScoringCheckpoint.from_json(checkpoint_str)
        except (TypeError, ValueError, json.JSONDecodeError) as e:
            _logger.debug(
                f"Failed to parse checkpoint for experiment {self._experiment_id}: {e}",
                exc_info=True,
            )

    def persist_checkpoint(self, checkpoint: OnlineTraceScoringCheckpoint) -> None:
        """
        Persist the checkpoint tag with a new checkpoint.

        Args:
            checkpoint: The checkpoint to store.
        """
        self._tracking_store.set_experiment_tag(
            self._experiment_id,
            ExperimentTag(MLFLOW_LATEST_ONLINE_SCORING_TRACE_CHECKPOINT, checkpoint.to_json()),
        )

    def calculate_time_window(self) -> OnlineTraceScoringTimeWindow:
        """
        Calculate the time window for trace scoring.

        Enforces a maximum lookback period to prevent getting stuck on persistently
        failing traces. If the checkpoint is older than MAX_LOOKBACK_MS, uses
        current_time - MAX_LOOKBACK_MS instead to skip over old problematic traces.

        Returns:
            OnlineTraceScoringTimeWindow with min and max trace timestamps.
            min_trace_timestamp_ms is the checkpoint if it exists and is within the
            lookback period, otherwise now - MAX_LOOKBACK_MS.
            max_trace_timestamp_ms is the current time.
        """
        current_time_ms = int(time.time() * 1000)
        checkpoint = self.get_checkpoint()

        # Start from checkpoint, but never look back more than MAX_LOOKBACK_MS
        min_lookback_time_ms = current_time_ms - MAX_LOOKBACK_MS

        if checkpoint is not None:
            min_trace_timestamp_ms = max(checkpoint.timestamp_ms, min_lookback_time_ms)
        else:
            min_trace_timestamp_ms = min_lookback_time_ms

        return OnlineTraceScoringTimeWindow(
            min_trace_timestamp_ms=min_trace_timestamp_ms,
            max_trace_timestamp_ms=current_time_ms,
        )



from dataclasses import dataclass
from enum import Enum

from mlflow.entities._mlflow_object import _MlflowObject
from mlflow.protos import service_pb2 as pb


class MetricViewType(str, Enum):
    TRACES = "TRACES"
    SPANS = "SPANS"
    ASSESSMENTS = "ASSESSMENTS"

    def __str__(self) -> str:
        return self.value

    def to_proto(self):
        return pb.MetricViewType.Value(self)

    @classmethod
    def from_proto(cls, proto: int) -> "MetricViewType":
        return cls(pb.MetricViewType.Name(proto))


class AggregationType(str, Enum):
    COUNT = "COUNT"
    SUM = "SUM"
    AVG = "AVG"
    PERCENTILE = "PERCENTILE"
    MIN = "MIN"
    MAX = "MAX"

    def __str__(self) -> str:
        return self.value

    def to_proto(self):
        return pb.AggregationType.Value(self)


@dataclass
class MetricAggregation(_MlflowObject):
    aggregation_type: AggregationType
    percentile_value: float | None = None

    def __post_init__(self):
        if self.aggregation_type == AggregationType.PERCENTILE:
            if self.percentile_value is None:
                raise ValueError("Percentile value is required for PERCENTILE aggregation")
            if self.percentile_value > 100 or self.percentile_value < 0:
                raise ValueError(
                    f"Percentile value must be between 0 and 100, got {self.percentile_value}"
                )
        elif self.percentile_value is not None:
            raise ValueError(
                "Percentile value is only allowed for PERCENTILE aggregation type, "
                f"got {self.aggregation_type}"
            )

    def __str__(self) -> str:
        if self.aggregation_type == AggregationType.PERCENTILE:
            return f"P{self.percentile_value}"
        return str(self.aggregation_type)

    def to_proto(self) -> pb.MetricAggregation:
        proto = pb.MetricAggregation()
        proto.aggregation_type = self.aggregation_type.to_proto()
        if self.percentile_value is not None:
            proto.percentile_value = self.percentile_value
        return proto

    @classmethod
    def from_proto(cls, proto: pb.MetricAggregation) -> "MetricAggregation":
        return cls(
            aggregation_type=AggregationType(pb.AggregationType.Name(proto.aggregation_type)),
            percentile_value=proto.percentile_value if proto.HasField("percentile_value") else None,
        )


@dataclass
class MetricDataPoint(_MlflowObject):
    metric_name: str
    dimensions: dict[str, str]
    values: dict[str, float]

    @classmethod
    def from_proto(cls, proto: pb.MetricDataPoint) -> "MetricDataPoint":
        return cls(
            metric_name=proto.metric_name,
            dimensions=dict(proto.dimensions),
            values=dict(proto.values),
        )

    def to_proto(self) -> pb.MetricDataPoint:
        return pb.MetricDataPoint(
            metric_name=self.metric_name,
            dimensions=self.dimensions,
            values=self.values,
        )



"""
Claude Code provider for MLflow Assistant.

This module provides the Claude Code integration for the assistant API,
enabling AI-powered trace analysis through the Claude Code CLI.
"""

import asyncio
import json
import logging
import shutil
import subprocess
from pathlib import Path
from typing import Any, AsyncGenerator, Callable

from mlflow.assistant.providers.base import (
    AssistantProvider,
    CLINotInstalledError,
    NotAuthenticatedError,
    load_config,
)
from mlflow.assistant.types import (
    ContentBlock,
    Event,
    Message,
    TextBlock,
    ThinkingBlock,
    ToolResultBlock,
    ToolUseBlock,
)

_logger = logging.getLogger(__name__)


# Allowed tools for Claude Code CLI
# Restrict to only Bash commands that use MLflow CLI
ALLOWED_TOOLS = [
    "Bash(mlflow:*)",
]

# TODO: to be updated
CLAUDE_SYSTEM_PROMPT = """You are an MLflow assistant helping users with their MLflow projects.

User messages may include a <context> block containing JSON that represents what the user is
currently viewing on screen (e.g., traceId, experimentId, selectedTraceIds). Use this context
to understand what entities the user is referring to when they ask questions.

## Command Preferences (IMPORTANT)

* When working with MLflow data and operations, ALWAYS use MLflow CLI commands directly.
* Never combine two bash command with `&&` or `||`. That will error out.
* Trust that MLflow CLI commands will work. Do not add error handling or fallbacks to Python.
* When using MLflow CLI, always use `--help` to discover all available
options. Do not skip this step or you will not get the correct command.
"""


class ClaudeCodeProvider(AssistantProvider):
    """Assistant provider using Claude Code CLI."""

    @property
    def name(self) -> str:
        return "claude_code"

    @property
    def display_name(self) -> str:
        return "Claude Code"

    @property
    def description(self) -> str:
        return "AI-powered assistant using Claude Code CLI"

    def is_available(self) -> bool:
        return shutil.which("claude") is not None

    def check_connection(self, echo: Callable[[str], None] | None = None) -> None:
        """
        Check if Claude CLI is installed and authenticated.

        Args:
            echo: Optional function to print status messages.

        Raises:
            ProviderNotConfiguredError: If CLI is not installed or not authenticated.
        """
        claude_path = shutil.which("claude")
        if not claude_path:
            if echo:
                echo("Claude CLI not found")
            raise CLINotInstalledError(
                "Claude Code CLI is not installed. "
                "Install it with: npm install -g @anthropic-ai/claude-code"
            )

        if echo:
            echo(f"Claude CLI found: {claude_path}")
            echo("Checking connection... (this may take a few seconds)")

        # Check authentication by running a minimal test prompt
        try:
            result = subprocess.run(
                ["claude", "-p", "hi", "--max-turns", "1", "--output-format", "json"],
                capture_output=True,
                text=True,
                timeout=30,
            )

            if result.returncode == 0:
                if echo:
                    echo("Authentication verified")
                return

            # Check for common auth errors in stderr
            stderr = result.stderr.lower()
            if "auth" in stderr or "login" in stderr or "unauthorized" in stderr:
                error_msg = "Not authenticated. Please run: claude login"
            else:
                error_msg = result.stderr.strip() or f"Process exited with code {result.returncode}"

            if echo:
                echo(f"Authentication failed: {error_msg}")
            raise NotAuthenticatedError(error_msg)

        except subprocess.TimeoutExpired:
            if echo:
                echo("Authentication check timed out")
            raise NotAuthenticatedError("Authentication check timed out")
        except subprocess.SubprocessError as e:
            if echo:
                echo(f"Error checking authentication: {e}")
            raise NotAuthenticatedError(str(e))

    def install_skills(self, skill_path: Path) -> list[str]:
        """Install MLflow-specific Claude skills.

        Args:
            skill_path: Directory where skills should be installed.
        """
        # Get the skills directory from this package
        skills_source = Path(__file__).parent.parent / "skills"

        if not skills_source.exists():
            raise RuntimeError("Skills directory not found")

        # Create destination directory
        skill_path.mkdir(parents=True, exist_ok=True)

        # Find all skill directories in the skills directory
        skill_dirs = [d for d in skills_source.iterdir() if d.is_dir()]
        if not skill_dirs:
            raise RuntimeError("No skills to install")

        installed_skills = []
        for skill_dir in skill_dirs:
            dest_skill_dir = skill_path / skill_dir.name
            dest_skill_dir.mkdir(parents=True, exist_ok=True)

            # Only copy files, not subdirectories, to avoid overwriting user customizations
            for file in skill_dir.iterdir():
                if file.is_file():
                    shutil.copy2(file, dest_skill_dir / file.name)

            installed_skills.append(skill_dir.name)

        return installed_skills

    async def astream(
        self,
        prompt: str,
        session_id: str | None = None,
        cwd: Path | None = None,
        context: dict[str, Any] | None = None,
    ) -> AsyncGenerator[Event, None]:
        """
        Stream responses from Claude Code CLI asynchronously.

        Args:
            prompt: The prompt to send to Claude
            session_id: Claude session ID for resume
            cwd: Working directory for Claude Code CLI
            context: Page context (experimentId, traceId, selectedTraceIds, etc.)

        Yields:
            Event objects
        """
        claude_path = shutil.which("claude")
        if not claude_path:
            yield Event.from_error(
                "Claude CLI not found. Please install Claude Code CLI and ensure it's in your PATH."
            )
            return

        # Build user message with context
        if context:
            user_message = f"<context>\n{json.dumps(context)}\n</context>\n\n{prompt}"
        else:
            user_message = prompt

        # Build command
        # Note: --verbose is required when using --output-format=stream-json with -p
        cmd = [claude_path, "-p", user_message, "--output-format", "stream-json", "--verbose"]

        # Add system prompt
        cmd.extend(["--append-system-prompt", CLAUDE_SYSTEM_PROMPT])

        # Add allowed tools restriction
        for tool in ALLOWED_TOOLS:
            cmd.extend(["--allowed-tools", tool])

        config = load_config(self.name)
        if config.model and config.model != "default":
            cmd.extend(["--model", config.model])

        if session_id:
            cmd.extend(["--resume", session_id])

        process = None
        try:
            process = await asyncio.create_subprocess_exec(
                *cmd,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
                cwd=cwd,
            )

            async for line in process.stdout:
                line_str = line.decode("utf-8").strip()
                if not line_str:
                    continue

                try:
                    data = json.loads(line_str)
                    if msg := self._parse_message_to_event(data):
                        yield msg

                except json.JSONDecodeError:
                    # Non-JSON output, treat as plain text
                    yield Event.from_message(Message(role="user", content=line_str))

            # Wait for process to complete
            await process.wait()

            if process.returncode != 0:
                stderr = await process.stderr.read()
                error_msg = (
                    stderr.decode("utf-8").strip()
                    or f"Process exited with code {process.returncode}"
                )
                yield Event.from_error(error_msg)

        except Exception as e:
            _logger.exception("Error running Claude Code CLI")
            yield Event.from_error(str(e))
        finally:
            if process is not None and process.returncode is None:
                process.kill()
                await process.wait()

    def _parse_message_to_event(self, data: dict[str, Any]) -> Event | None:
        """
        Parse json message from Claude Code CLI output.

        Reference: https://github.com/anthropics/claude-agent-sdk-python/blob/29c12cd80b256e88f321b2b8f1f5a88445077aa5/src/claude_agent_sdk/_internal/message_parser.py#L24

        Args:
            data: Raw message dictionary from CLI output

        Returns:
            Parsed Event object
        """
        message_type = data.get("type")
        if not message_type:
            return Event.from_error("Message missing 'type' field")

        match message_type:
            case "user":
                try:
                    if isinstance(data["message"]["content"], list):
                        user_content_blocks = []
                        for block in data["message"]["content"]:
                            match block["type"]:
                                case "text":
                                    user_content_blocks.append(TextBlock(text=block["text"]))
                                case "tool_use":
                                    user_content_blocks.append(
                                        ToolUseBlock(
                                            id=block["id"],
                                            name=block["name"],
                                            input=block["input"],
                                        )
                                    )
                                case "tool_result":
                                    user_content_blocks.append(
                                        ToolResultBlock(
                                            tool_use_id=block["tool_use_id"],
                                            content=block.get("content"),
                                            is_error=block.get("is_error"),
                                        )
                                    )
                            msg = Message(role="user", content=user_content_blocks)
                    else:
                        msg = Message(role="user", content=data["message"]["content"])
                    return Event.from_message(msg)
                except KeyError as e:
                    return Event.from_error(f"Failed to parse user message: {e}")

            case "assistant":
                try:
                    if data["message"].get("error"):
                        return Event.from_error(data["message"]["error"])

                    content_blocks: list[ContentBlock] = []
                    for block in data["message"]["content"]:
                        match block["type"]:
                            case "text":
                                content_blocks.append(TextBlock(text=block["text"]))
                            case "thinking":
                                content_blocks.append(
                                    ThinkingBlock(
                                        thinking=block["thinking"],
                                        signature=block["signature"],
                                    )
                                )
                            case "tool_use":
                                content_blocks.append(
                                    ToolUseBlock(
                                        id=block["id"],
                                        name=block["name"],
                                        input=block["input"],
                                    )
                                )
                            case "tool_result":
                                content_blocks.append(
                                    ToolResultBlock(
                                        tool_use_id=block["tool_use_id"],
                                        content=block.get("content"),
                                        is_error=block.get("is_error"),
                                    )
                                )

                    msg = Message(role="assistant", content=content_blocks)
                    return Event.from_message(msg)
                except KeyError as e:
                    return Event.from_error(f"Failed to parse assistant message: {e}")

            case "system":
                # NB: Skip system message. The system message from Claude Code CLI contains
                # the various metadata about runtime, which is not used by the assistant UX.
                return None

            case "error":
                try:
                    error_msg = data.get("error", {}).get("message", str(data.get("error")))
                    return Event.from_error(error_msg)
                except Exception as e:
                    return Event.from_error(f"Failed to parse error message: {e}")

            case "result":
                try:
                    return Event.from_result(
                        result=data.get("result"),
                        session_id=data["session_id"],
                    )
                except KeyError as e:
                    return Event.from_error(f"Failed to parse result message: {e}")

            case "stream_event":
                try:
                    return Event.from_stream_event(event=data["event"])
                except KeyError as e:
                    return Event.from_error(f"Failed to parse stream_event message: {e}")

            case _:
                return Event.from_error(f"Unknown message type: {message_type}")
